%{
#include <stdlib.h>
#include <string.h>

#include "script.h"

#define YYERROR_VERBOSE

int yylex(void);

void yyerror(const char* s) {
	script_error(s);
}

struct script_cmd* yyresult;

%}

%union {
	int val;
	const char* str;
	struct script_exp* exp;
	struct script_cmd* cmd;
}

%token_table

%left OP_LAND OP_LOR
%left '&' '|' '^'
%left '<' '>' OP_E OP_LE OP_GE
%left OP_SL OP_SR
%left '+' '-'
%left '~' '!'

%token WAIT
%token DELAY
%token <val> VAL
%token <str> STRING
%type <cmd> cmd cmd_list script
%type <exp> exp

%%

script: cmd_list {
		yyresult = $1;
		YYACCEPT;
	}

cmd_list: cmd_list cmd {
		$$ = script_cmd_make_op2cc($1, $2);
		if (!$$)
			YYERROR;
	}
	| cmd {
		$$ = $1; 
	}
;

exp:	VAL {
		$$ = script_exp_make_op1v(SCRIPT_EXP_VALUE,$1);
		if (!$$) YYERROR;
	}
	| STRING {
		$$ = script_exp_make_op1s(SCRIPT_EXP_VARIABLE,$1);
		if (!$$) YYERROR;
	}
	| '(' exp ')' {
		$$ = $2;
	}
	| exp '+' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_ADD,$1,$3);
		if (!$$) YYERROR;
	}
	| exp '-' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_SUB,$1,$3);
		if (!$$) YYERROR;
	}
	| exp '&' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_AND,$1,$3);
		if (!$$) YYERROR;
	}
	| exp '|' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_OR,$1,$3);
		if (!$$) YYERROR;
	}
	| exp '^' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_XOR,$1,$3);
		if (!$$) YYERROR;
	}
	| exp '<' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_L,$1,$3);
		if (!$$) YYERROR;
	}
	| exp '>' exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_G,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_E exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_E,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_LE exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_LE,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_GE exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_GE,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_SL exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_SL,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_SR exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_SR,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_LAND exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_LAND,$1,$3);
		if (!$$) YYERROR;
	}
	| exp OP_LOR exp {
		$$ = script_exp_make_op2ee(SCRIPT_EXP_LOR,$1,$3);
		if (!$$) YYERROR;
	}
	| '!' exp {
		$$ = script_exp_make_op1e(SCRIPT_EXP_LNOT,$2);
		if (!$$) YYERROR;
	}
	| '~' exp {
		$$ = script_exp_make_op1e(SCRIPT_EXP_NOT,$2);
		if (!$$) YYERROR;
	}
	| STRING '(' ')' {
		$$ = script_exp_make_op1f(SCRIPT_EXP_F0,$1);
		if (!$$) YYERROR;
	}
	| STRING '(' exp ')' {
		$$ = script_exp_make_op2fe(SCRIPT_EXP_F1,$1,$3);
		if (!$$) YYERROR;
	}
	| STRING '(' exp ',' exp ')' {
		$$ = script_exp_make_op3fee(SCRIPT_EXP_F2,$1,$3,$5);
		if (!$$) YYERROR;
	}
;

cmd:	'{' cmd_list '}' {
		$$ = script_cmd_make_op2sc("inner",$2);
		if (!$$) YYERROR;
	}
	| STRING '(' exp ')' '{' cmd_list '}' {
		$$ = script_cmd_make_op3sec($1,$3,$6);
		if (!$$) YYERROR;
	}
	| STRING '{' cmd_list '}' {
		$$ = script_cmd_make_op2sc($1,$3);
		if (!$$) YYERROR;
	}
	| WAIT '(' exp ')' ';' {
		$$ = script_cmd_make_op2se("wait",$3);
		if (!$$) YYERROR;
	}
	| DELAY '(' exp ')' ';' {
		$$ = script_cmd_make_op2se("delay",$3);
		if (!$$) YYERROR;
	}
	| exp ';' {
		$$ = script_cmd_make_op2se("evaluate",$1);
		if (!$$) YYERROR;
	}
;

%%

const char* script_input_begin;
const char* script_input_end;

int script_input(char* buf, int max_size) {
	(void)max_size;
	if (script_input_begin == script_input_end) {
		return 0;
	} else {
		*buf = *script_input_begin++;
		return 1;
	}
}

struct script_cmd* script_parse(const char* text) {
	int r;
	script_input_begin = text;
	script_input_end = script_input_begin + strlen(script_input_begin);
	r = yyparse();
	if (r!=0)
		return 0;
	r = yylex();
	if (r!=0) {
		if (r == STRING) {
			char buffer[256];
			sprintf(buffer,"Unexpected token '%s'",yylval.str);
			yyerror(buffer);
		} else if (r<255) {
			char buffer[256];
			sprintf(buffer,"Unexpected token '%c'",(char)r);
			yyerror(buffer);
		} else {
			yyerror("Unexpected token");
		}
		return 0;
	}
	return yyresult;
}
