/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "generate.h"
#include "vbeline.h"

#include <math.h>

static unsigned step(double v, unsigned st) {
	unsigned il,ih;
	double vl = floor(v);
	il = (unsigned)vl;
	il -= il % st;
	ih = il + st;
	if (fabs(v - il) / fabs(v - ih) < (1.0 + 1e-6) ) // + favorite the lower value
	return il;
	else
	return ih;
}

/* Allowed steps in CRTC values */
#define HSTEP 8
#define VSTEP 1

/* Normalize the values */
void generate_normalize(video_generate* generate) {
	double ht = generate->hactive + generate->hfront +  generate->hsync + generate->hback;
	double vt = generate->vactive + generate->vfront +  generate->vsync + generate->vback;
	generate->hactive /= ht;
	generate->hfront /= ht;
	generate->hsync /= ht;
	generate->hback /= ht;
	generate->vactive /= vt;
	generate->vfront /= vt;
	generate->vsync /= vt;
	generate->vback /= vt;
}

/* Generate the CRTC values */
void generate_crtc(video_crtc* crtc, unsigned hsize, unsigned vsize, const video_generate* generate) {
	video_generate norm = *generate;

	/* use normalized values */
	generate_normalize(&norm);

	crtc->hde = step(hsize, HSTEP);
	crtc->ht = step(crtc->hde / norm.hactive, HSTEP);
	crtc->hrs = step(crtc->ht * (norm.hactive + norm.hfront), HSTEP);
	crtc->hre = step(crtc->ht * (norm.hactive + norm.hfront + norm.hsync), HSTEP);
	if (crtc->hrs == crtc->hre)
		crtc->hre = crtc->hrs + HSTEP;

	crtc->vde = step(vsize, VSTEP);
	crtc->vt = step(crtc->vde / norm.vactive, VSTEP);
	crtc->vrs = step(crtc->vt * (norm.vactive + norm.vfront), VSTEP);
	crtc->vre = step(crtc->vt * (norm.vactive + norm.vfront + norm.vsync), VSTEP);
	if (crtc->vrs == crtc->vre)
		crtc->vre = crtc->vrs + VSTEP;
}

/* Set the pixel clock */
void generate_pixel_clock(video_crtc* crtc, double clock) {
	crtc->pixelclock = clock;
}

/* Set the horz clock */
void generate_horz_clock(video_crtc* crtc, double clock) {
	crtc->pixelclock = clock * crtc->ht;
}

/* Set the vert clock */
void generate_vert_clock(video_crtc* crtc, double clock) {
	unsigned factor = 1;
	if (!crtc->interlaced)
		factor *= 2;
	if (crtc->doublescan)
		factor *= 2;
	crtc->pixelclock = clock * crtc->ht * crtc->vt * factor / 2;
}

/* Set default value for a VGA monitor */
void generate_default_vga(video_generate* generate) {
	generate->hactive = 640;
	generate->hfront = 16;
	generate->hsync = 96;
	generate->hback = 48;
	generate->vactive = 480;
	generate->vfront = 10;
	generate->vsync = 2;
	generate->vback = 33;
}

void generate_default_atari_standard(video_generate* generate) {
	generate->hactive = 46.9;
	generate->hfront = 4.7;
	generate->hsync = 11.9;
	generate->hback = 63.6;
	generate->vactive = 15.3;
	generate->vfront = 0.2;
	generate->vsync = 1.2;
	generate->vback = 16.7;
}

void generate_default_atari_extended(video_generate* generate) {
	generate->hactive = 48;
	generate->hfront = 3.9;
	generate->hsync = 11.9;
	generate->hback = 60.6;
	generate->vactive = 17.4;
	generate->vfront = 0.2;
	generate->vsync = 1.2;
	generate->vback = 18.9;
}

void generate_default_atari_medium(video_generate* generate) {
	generate->hactive = 32;
	generate->hfront = 4;
	generate->hsync = 7.2;
	generate->hback = 40;
	generate->vactive = 15.4;
	generate->vfront = 0.2;
	generate->vsync = 1.2;
	generate->vback = 16.7;
}

void generate_default_atari_vga(video_generate* generate) {
	generate->hactive = 25.6;
	generate->hfront = 4;
	generate->hsync = 5.7;
	generate->hback = 31.7;
	generate->vactive = 12.2;
	generate->vfront = 0.2;
	generate->vsync = 1.1;
	generate->vback = 14.3;
}

void generate_default_pal(video_generate* generate) {
	generate->hactive = 52.00;
	generate->hfront = 1.65;
	generate->hsync = 4.70;
	generate->hback = 5.65;
	generate->vactive = 288.5;
	generate->vfront = 3;
	generate->vsync = 3;
	generate->vback = 18;
}

void generate_default_ntsc(video_generate* generate) {
	generate->hactive = 52.60;
	generate->hfront = 1.50;
	generate->hsync = 4.70;
	generate->hback = 4.70;
	generate->vactive = 242.5;
	generate->vfront = 3;
	generate->vsync = 3;
	generate->vback = 14;
}

/* Adjust the clock to match the monitor specifications
	Return 0 if successful
*/
video_error generate_adjust_clock(video_crtc* crtc, const video_monitor* monitor) {
	double hclock;
	double vclock;
	double factor;
	double best_hclock = 0;
	int best_hclock_found = 0;
	int i;

	factor = 1;
	if (crtc->interlaced)
		factor *= 2;
	if (crtc->doublescan)
		factor /= 2;

	hclock = (double)crtc->pixelclock / crtc->ht;
	vclock = hclock / crtc->vt * factor;

	/* check if already valid */
	if (video_monitor_clock_check(monitor, hclock, vclock))
		return 0;

	/* check for every hclock limit */
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
		if (monitor->hclock[i].low != 0 && monitor->hclock[i].high != 0) {
			double try_hclock, try_vclock;

			/* low limit */
			try_hclock = monitor->hclock[i].low;
			try_vclock = try_hclock / crtc->vt * factor;
			if (video_monitor_vclock_check(monitor, try_vclock)) {
				if (!best_hclock_found || fabs(try_hclock - hclock) < fabs(best_hclock - hclock)) {
					best_hclock = try_hclock;
					best_hclock_found = 1;
				}
			}

			/* high limit */
			try_hclock = monitor->hclock[i].high;
			try_vclock = try_hclock / crtc->vt * factor;
			if (video_monitor_vclock_check(monitor, try_vclock)) {
				if (!best_hclock_found || fabs(try_hclock - hclock) < fabs(best_hclock - hclock)) {
					best_hclock = try_hclock;
					best_hclock_found = 1;
				}
			}
		}
	}

	/* check for every vclock limit */
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
		if (monitor->vclock[i].low != 0 && monitor->vclock[i].high != 0) {
			double try_hclock, try_vclock;

			/* low limit */
			try_vclock = monitor->vclock[i].low;
			try_hclock = try_vclock * crtc->vt / factor;
			if (video_monitor_hclock_check(monitor, try_hclock)) {
				if (!best_hclock_found || fabs(try_hclock - hclock) < fabs(best_hclock - hclock)) {
					best_hclock = try_hclock;
					best_hclock_found = 1;
				}
			}

			/* high limit */
			try_vclock = monitor->vclock[i].high;
			try_hclock = try_vclock * crtc->vt / factor;
			if (video_monitor_hclock_check(monitor, try_hclock)) {
				if (!best_hclock_found || fabs(try_hclock - hclock) < fabs(best_hclock - hclock)) {
					best_hclock = try_hclock;
					best_hclock_found = 1;
				}
			}
		}
	}

	if (best_hclock_found) {
		crtc->pixelclock = best_hclock * crtc->ht;
		return 0;
	} else {
		return -1;
	}
}

video_error generate_find_exact(unsigned req_vtotal, double req_vclock, const video_monitor* monitor) {
	if (!video_monitor_clock_check(monitor, req_vclock * req_vtotal, req_vclock))
		return -1;
	return 0;
}

/* Check the monitor specification for the nearest vclock available */
static video_error generate_find_nearest_vclock(unsigned req_vtotal, double* req_vclock, const video_monitor* monitor) {
	double vclock;
	double best_vclock = 0;
	int best_found = 0;
	int i;

	vclock = *req_vclock;

	/* check for every hclock limit */
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
		if (monitor->hclock[i].low != 0 && monitor->hclock[i].high != 0) {
			double try_hclock, try_vclock;

			/* low limit */
			try_hclock = monitor->hclock[i].low;
			try_vclock = try_hclock / req_vtotal;
			if (video_monitor_vclock_check(monitor, try_vclock)) {
				if (!best_found || fabs(try_vclock - vclock) < fabs(best_vclock - vclock)) {
					best_vclock = try_vclock;
					best_found = 1;
				}
			}

			/* high limit */
			try_hclock = monitor->hclock[i].high;
			try_vclock = try_hclock / req_vtotal;
			if (video_monitor_vclock_check(monitor, try_vclock)) {
				if (!best_found || fabs(try_vclock - vclock) < fabs(best_vclock - vclock)) {
					best_vclock = try_vclock;
					best_found = 1;
				}
			}
		}
	}

	/* check for every vclock limit */
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
		if (monitor->vclock[i].low != 0 && monitor->vclock[i].high != 0) {
			double try_hclock, try_vclock;

			/* low limit */
			try_vclock = monitor->vclock[i].low;
			try_hclock = try_vclock * req_vtotal;
			if (video_monitor_hclock_check(monitor, try_hclock)) {
				if (!best_found || fabs(try_vclock - vclock) < fabs(best_vclock - vclock)) {
					best_vclock = try_vclock;
					best_found = 1;
				}
			}

			/* high limit */
			try_vclock = monitor->vclock[i].high;
			try_hclock = try_vclock * req_vtotal;
			if (video_monitor_hclock_check(monitor, try_hclock)) {
				if (!best_found || fabs(try_vclock - vclock) < fabs(best_vclock - vclock)) {
					best_vclock = try_vclock;
					best_found = 1;
				}
			}
		}
	}

	if (!best_found)
		return -1;

	*req_vclock = best_vclock;
	return 0;
}

/* Check the monitor specification for the nearest vtotal available */
static video_error generate_find_nearest_vtotal(unsigned* req_vtotal, double* req_vclock, const video_monitor* monitor) {
	unsigned vtotal;
	unsigned best_vtotal = 0;
	double best_vclock = 0;
	int best_found = 0;
	int i,j;

	vtotal = *req_vtotal;

	/* check for every hclock limit */
	for(j=0;j<VIDEO_MONITOR_RANGE_MAX;++j) {
		if (monitor->hclock[j].low != 0 && monitor->hclock[j].high != 0) {
 			/* check for every vclock limit */
			for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
				if (monitor->vclock[i].low != 0 && monitor->vclock[i].high != 0) {
					unsigned try_vtotal;
					double try_hclock;
					double try_vclock;

					/* low limit */
					/* the floor() approximation uses a greater vclock */
					try_vtotal = floor(monitor->hclock[j].high / monitor->vclock[i].low);
					try_hclock = monitor->hclock[j].high;
					try_vclock = try_hclock / try_vtotal;
					if (video_monitor_clock_check(monitor, try_hclock, try_vclock)) {
						if (!best_found || fabs(try_vtotal - vtotal) < fabs(best_vtotal - vtotal)) {
							best_vtotal = try_vtotal;
							best_vclock = try_vclock;
							best_found = 1;
						}
					}

					/* high limit */
					/* the ceil() approximation uses a lesser vclock */
					try_vtotal = ceil(monitor->hclock[j].low / monitor->vclock[i].high);
					try_hclock = monitor->hclock[j].low;
					try_vclock = try_hclock / try_vtotal;
					if (video_monitor_clock_check(monitor, try_hclock, try_vclock)) {
						if (!best_found || fabs(try_vtotal - vtotal) < fabs(best_vtotal - vtotal)) {
							best_vtotal = try_vtotal;
							best_vclock = try_vclock;
							best_found = 1;
						}
					}
				}
			}
		}
	}

	if (!best_found)
		return -1;

	*req_vtotal = best_vtotal;
	*req_vclock = best_vclock;
	return 0;
}

/* Find the nearest vertical size available */
static video_error generate_find_best(unsigned* req_vtotal, double* req_vclock, double* req_factor, const video_monitor* monitor, unsigned cap) {
	int best_found = 0;
	double best_vclock = 0;
	double best_factor = 0;
	unsigned best_vtotal = 0;

	/* exact match */
	if ((cap & VBELINE_CAP_SINGLESCAN) != 0) {
		if (generate_find_exact(*req_vtotal,*req_vclock,monitor)==0) {
			*req_factor = 1;
			return 0;
		}
	}

	if ((cap & VBELINE_CAP_DOUBLESCAN) != 0) {
		if (generate_find_exact(*req_vtotal * 2,*req_vclock,monitor)==0) {
			*req_factor = 2;
			return 0;
		}
	}

	if ((cap & VBELINE_CAP_INTERLACE) != 0) {
		if (generate_find_exact(*req_vtotal / 2,*req_vclock,monitor)==0) {
			*req_factor = 0.5;
			return 0;
		}
	}

	/* best vclock */
	if ((cap & VBELINE_CAP_SINGLESCAN) != 0) {
		double try_vclock = *req_vclock;
		if (generate_find_nearest_vclock(*req_vtotal,&try_vclock,monitor)==0) {
			if (!best_found || fabs(try_vclock - *req_vclock) < fabs(best_vclock - *req_vclock)) {
				best_factor = 1;
			 	best_vclock = try_vclock;
				best_found = 1;
			}
		}
	}

	if ((cap & VBELINE_CAP_DOUBLESCAN) != 0) {
		double try_vclock = *req_vclock;
		if (generate_find_nearest_vclock(*req_vtotal * 2,&try_vclock,monitor)==0) {
			if (!best_found || fabs(try_vclock - *req_vclock) < fabs(best_vclock - *req_vclock)) {
				best_factor = 2;
			 	best_vclock = try_vclock;
				best_found = 1;
			}
		}
	}

	if ((cap & VBELINE_CAP_INTERLACE) != 0) {
		double try_vclock = *req_vclock;
		if (generate_find_nearest_vclock(*req_vtotal / 2,&try_vclock,monitor)==0) {
			if (!best_found || fabs(try_vclock - *req_vclock) < fabs(best_vclock - *req_vclock)) {
				best_factor = 2;
			 	best_vclock = try_vclock;
				best_found = 1;
			}
		}
	}

	if (best_found) {
		*req_factor = best_factor;
		*req_vclock = best_vclock;
		return 0;
	}

	/* best vtotal */
	if ((cap & VBELINE_CAP_SINGLESCAN) != 0) {
		double try_vclock = *req_vclock;
		unsigned try_vtotal = *req_vtotal;
		if (generate_find_nearest_vtotal(&try_vtotal,&try_vclock,monitor)==0) {
			if (!best_found || fabs(try_vtotal - *req_vtotal) < fabs(best_vtotal - *req_vtotal)) {
				best_factor = 1;
			 	best_vclock = try_vclock;
				best_vtotal = try_vtotal;
				best_found = 1;
			}
		}
	}

	if ((cap & VBELINE_CAP_DOUBLESCAN) != 0) {
		double try_vclock = *req_vclock;
		unsigned try_vtotal = *req_vtotal * 2;
		if (generate_find_nearest_vtotal(&try_vtotal,&try_vclock,monitor)==0) {
			if (!best_found || fabs(try_vtotal - *req_vtotal) < fabs(best_vtotal - *req_vtotal)) {
				best_factor = 2;
			 	best_vclock = try_vclock;
				best_vtotal = try_vtotal / 2;
				best_found = 1;
			}
		}
	}

	if ((cap & VBELINE_CAP_INTERLACE) != 0) {
		double try_vclock = *req_vclock;
		unsigned try_vtotal = *req_vtotal / 2;
		if (generate_find_nearest_vtotal(&try_vtotal,&try_vclock,monitor)==0) {
			if (!best_found || fabs(try_vtotal - *req_vtotal) < fabs(best_vtotal - *req_vtotal)) {
				best_factor = 0.5;
			 	best_vclock = try_vclock;
				best_vtotal = try_vtotal * 2;
				best_found = 1;
			}
		}
	}

	if (best_found) {
		*req_factor = best_factor;
		*req_vclock = best_vclock;
		*req_vtotal = best_vtotal;
		return 0;
	}

	return -1;
}

video_error generate_find(video_crtc* crtc, unsigned hsize, unsigned vsize, double vclock, const video_monitor* monitor, const video_generate* generate, unsigned capability) {
	video_generate norm = *generate;
	unsigned vtotal;
	double factor;

	/* use normalized values */
	generate_normalize(&norm);

	/* required */
	vtotal = step(vsize / norm.vactive, VSTEP);;
	factor = 0;

	if (generate_find_best(&vtotal, &vclock, &factor, monitor, capability)!=0)
		return -1;

	/* doublescan/interlace */
	if (factor == 1.0) {
		crtc->doublescan = 0;
		crtc->interlaced = 0;
	} else if (factor == 2.0) {
		crtc->doublescan = 1;
		crtc->interlaced = 0;
	} else if (factor == 0.5) {
		crtc->doublescan = 0;
		crtc->interlaced = 1;
	} else
		return -1;

	/* polarization */
	crtc->hsync = 1;
	crtc->vsync = 1;

	/* compute the horizontal crtc */
	crtc->hde = step(hsize, HSTEP);
	crtc->ht = step(crtc->hde / norm.hactive, HSTEP);
	crtc->hrs = step(crtc->ht * (norm.hactive + norm.hfront), HSTEP);
	crtc->hre = step(crtc->ht * (norm.hactive + norm.hfront + norm.hsync), HSTEP);
	if (crtc->hrs == crtc->hre)
		crtc->hre = crtc->hrs + HSTEP;

	/* compute the vertical crtc */
	crtc->vt = step(vtotal, VSTEP);
	crtc->vde = step(crtc->vt * norm.vactive, VSTEP);
	if (abs(crtc->vde - vsize) <= VSTEP) /* solve precision problem */
		crtc->vde = step(vsize, VSTEP);
	crtc->vrs = step(crtc->vt * (norm.vactive + norm.vfront), VSTEP);
	crtc->vre = step(crtc->vt * (norm.vactive + norm.vfront + norm.vsync), VSTEP);
	if (crtc->vrs == crtc->vre)
		crtc->vre = crtc->vrs + VSTEP;

	/* pixelclock */
	crtc->pixelclock = vclock * factor * (crtc->vt * crtc->ht);

	/* final check, this should never fail */
	if (!video_monitor_crtc_check(monitor,crtc))
		return -1;

	return 0;
}

void generate_linear_interpolate(video_generate* monitor, unsigned hsize, unsigned vsize, const video_generate_linear* generate_linear) {
	double h = (hsize - (double)generate_linear->hsize0) / (generate_linear->hsize1 - (double)generate_linear->hsize0);
	double v = (vsize - (double)generate_linear->vsize0) / (generate_linear->vsize1 - (double)generate_linear->vsize0);

	monitor->hactive = generate_linear->gen0.hactive * (1-h) + generate_linear->gen1.hactive * h;
	monitor->hfront = generate_linear->gen0.hfront * (1-h) + generate_linear->gen1.hfront * h;
	monitor->hsync = generate_linear->gen0.hsync * (1-h) + generate_linear->gen1.hsync * h;
	monitor->hback = generate_linear->gen0.hback * (1-h) + generate_linear->gen1.hback * h;

	monitor->vactive = generate_linear->gen0.vactive * (1-v) + generate_linear->gen1.vactive * v;
	monitor->vfront = generate_linear->gen0.vfront * (1-v) + generate_linear->gen1.vfront * v;
	monitor->vsync = generate_linear->gen0.vsync * (1-v) + generate_linear->gen1.vsync * v;
	monitor->vback = generate_linear->gen0.vback * (1-v) + generate_linear->gen1.vback * v;
}

