/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __SCREEN_H
#define __SCREEN_H

#ifdef __cplusplus
extern "C" {
#endif

#include "scrstd.h"
#include "scrtweak.h"
#include "scrvga.h"
#include "scrvbe.h"

#include <string.h>
#include <go32.h>

/* Code used for check the result of operation
 * ==0 is ok
 * !=0 is not ok
 */
typedef int video_error;

/***************************************************************************/
/* Video CRTC */

typedef struct video_crtc_struct {
	unsigned hde,hrs,hre,ht; /* Normalized 1 unit==1 pixel */
	unsigned vde,vrs,vre,vt; /* Normalized 1 unit==1 pixel (also for doublescan and interlaced modes) */
	unsigned pixelclock; /* Pixel clock, (ignore doublescan and interlaced) */
	/* (this imply that this value is doubled if doublescan is active, or halved if interlaced) */
	int doublescan; /* Double scan mode 0==no,1==yes */
	int vsync; /* Horizontal sync polarity 1==negative,0==positive */
	int hsync; /* Horizontal sync polarity 1==negative,0==positive */
	int interlaced; /* Interlaced mode 0==no,1==yes */
} video_crtc;

/***************************************************************************/
/* Video mode */

/* Max length of video mode name */
#define VIDEO_MODE_NAME_MAX 64

typedef struct vbe_video_mode_struct {
	unsigned mode; /* VBE mode number (with | vbeLinearBuffer if used) */
} vbe_video_mode;

typedef struct vbeline_video_mode_struct {
	unsigned mode; /* VBE mode number (with | vbeLinearBuffer if used) */
	video_crtc crtc;
} vbeline_video_mode;

typedef struct vga_video_mode_struct {
	unsigned mode; /* VGA mode number */
} vga_video_mode;

/* Max number of regs used for tweak video mode */
#define VIDEO_MODE_REGS_MAX 64

typedef struct vgatweak_video_mode_struct {
	unsigned mode; /* VGA mode number */
	unsigned regs_num;
	tweak_register regs[VIDEO_MODE_REGS_MAX];
} vgatweak_video_mode;

typedef struct vgaline_video_mode_struct {
	unsigned mode; /* VGA mode number */
	video_crtc crtc;
} vgaline_video_mode;

/* Flags for video mode */
#define VIDEO_FLAGS_DRIVER_VGA 0x0001
#define VIDEO_FLAGS_DRIVER_VGALINE 0x0002
#define VIDEO_FLAGS_DRIVER_VGATWEAK 0x0003
#define VIDEO_FLAGS_DRIVER_VBE 0x0004
#define VIDEO_FLAGS_DRIVER_VBELINE 0x0005
#define VIDEO_FLAGS_DRIVER_MASK 0x000F

/* Flags for linear acces of memory support */
#define VIDEO_FLAGS_LINEAR 0x0010

/* Flags for triple buffer support */
#define VIDEO_FLAGS_SYNC_SETPAGE 0x0020

/* Flags for double buffer support */
#define VIDEO_FLAGS_ASYNC_SETPAGE 0x0040

/* Flags for color mode */
#define VIDEO_FLAGS_INDEX_RGB 0x0000 /* RGB mode */
#define VIDEO_FLAGS_INDEX_PACKED 0x0100 /* Palette mode */
#define VIDEO_FLAGS_INDEX_TEXT 0x0200 /* Text mode */
#define VIDEO_FLAGS_INDEX_MASK 0x0300

/* Flags for text or graphics mode */
#define VIDEO_FLAGS_TYPE_TEXT 0x0000 /* Text mode */
#define VIDEO_FLAGS_TYPE_GRAPHICS 0x0400 /* Graphics mode */
#define VIDEO_FLAGS_TYPE_MASK 0x0400

/* Users flags */
#define VIDEO_FLAGS_USER_BIT0 0x1000
#define VIDEO_FLAGS_USER_BIT1 0x2000
#define VIDEO_FLAGS_USER_BIT2 0x4000
#define VIDEO_FLAGS_USER_BIT3 0x8000
#define VIDEO_FLAGS_USER_MASK 0xF000

typedef struct video_mode_struct {
	unsigned flags;
	char name[VIDEO_MODE_NAME_MAX];
	unsigned size_x;
	unsigned size_y;
	unsigned virtual_x;
	unsigned virtual_y;
	double vert_freq; /* vertical freq, ==0 if not computable */
	double horz_freq; /* horiz freq, ==0 if not computable */
	unsigned bits; /* 8,15,16 */
	unsigned aspect_x;
	unsigned aspect_y;

	union {
		vga_video_mode vga;
		vgatweak_video_mode vgatweak;
		vgaline_video_mode vgaline;
		vbe_video_mode vbe;
		vbeline_video_mode vbeline;
	} driver;

	struct video_mode_struct* next; /* used for container */
} video_mode;

/***************************************************************************/
/* CRTC */

void video_crtc_blank_reset(video_crtc* crtc);
int video_option_crtc_blank_get(void);

video_error video_crtc_import(video_crtc* crtc, struct tweak_crtc* info, video_mode* mode, double vert_freq);
void video_crtc_realize_h(const video_crtc* crtc, unsigned *de, unsigned* bs, unsigned* rs, unsigned* re, unsigned* be, unsigned* tt, unsigned x_div);
void video_crtc_realize_v(const video_crtc* crtc, unsigned *de, unsigned* bs, unsigned* rs, unsigned* re, unsigned* be, unsigned* tt, unsigned y_mul);

/***************************************************************************/

/* Interface for twkuser */
unsigned tweak_register_get(unsigned port, int port_index, void* arg);
int tweak_register_set(unsigned value, unsigned port, int port_index, void* arg);

/* Name of the video mode */
static __inline__  const char* video_mode_name(const video_mode* mode) {
	return mode->name;
}

static __inline__  double video_mode_vert_freq(const video_mode* mode) {
	return mode->vert_freq;
}

static __inline__  double video_mode_horz_freq(const video_mode* mode) {
	return mode->horz_freq;
}

static __inline__  unsigned video_mode_size_x(const video_mode* mode) {
	return mode->size_x;
}

static __inline__  unsigned video_mode_size_y(const video_mode* mode) {
	return mode->size_y;
}

static __inline__  unsigned video_mode_virtual_x(const video_mode* mode) {
	return mode->virtual_x;
}

static __inline__  unsigned video_mode_virtual_y(const video_mode* mode) {
	return mode->virtual_y;
}

static __inline__  unsigned video_mode_bits_per_pixel(const video_mode* mode) {
	return mode->bits;
}

static __inline__  unsigned video_mode_bytes_per_pixel(const video_mode* mode) {
	return (video_mode_bits_per_pixel(mode) + 7) / 8;
}

static __inline__  int video_mode_flags(const video_mode* mode) {
	return mode->flags;
}

static __inline__  int video_mode_driver(const video_mode* mode) {
	return video_mode_flags(mode) & VIDEO_FLAGS_DRIVER_MASK;
}

static __inline__  int video_mode_is_text(const video_mode* mode) {
	return (video_mode_flags(mode) & VIDEO_FLAGS_TYPE_MASK) == VIDEO_FLAGS_TYPE_TEXT;
}

static __inline__  int video_mode_is_graphics(const video_mode* mode) {
	return (video_mode_flags(mode) & VIDEO_FLAGS_TYPE_MASK) == VIDEO_FLAGS_TYPE_GRAPHICS;
}

static __inline__  int video_mode_index(const video_mode* mode) {
	return video_mode_flags(mode) & VIDEO_FLAGS_INDEX_MASK;
}

/***************************************************************************/
/* Option */

video_error video_options_load(const char* (*get_string)(const char *section, const char *name, const char *def));

/* Select the blank range computation mode */
#define VIDEO_OPTION_CRTC_BLANK_MODERATE 1 /* Blank the least possible */
#define VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE 2 /* Blank in a reasonable manner (default) */
void video_option_crtc_blank_set(int mode);
int video_option_crtc_blank_is(void);

/* Select the vbeline tweaking mode */
#define VIDEO_OPTION_VBELINE_TWEAK_MODERATE 0 /* Modify only Display End and Blank registers */
#define VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE 1 /* Rewrite all registers (default) */
void video_option_vbeline_tweak_set(int mode);
int video_option_vbeline_tweak_get(void);

/* Select the vbeline driver use */
#define VIDEO_OPTION_VBELINE_DRIVER_NONE 0 /* Doesn't use drivers */
#define VIDEO_OPTION_VBELINE_DRIVER_AUTO 1 /* (default) */
#define VIDEO_OPTION_VBELINE_DRIVER_VBE3 2
#define VIDEO_OPTION_VBELINE_DRIVER_LAGUNA 4
#define VIDEO_OPTION_VBELINE_DRIVER_SIS6326 5
#define VIDEO_OPTION_VBELINE_DRIVER_SAVAGE 6
#define VIDEO_OPTION_VBELINE_DRIVER_3DFX 7
#define VIDEO_OPTION_VBELINE_DRIVER_ATI 8
#define VIDEO_OPTION_VBELINE_DRIVER_S3 9
#define VIDEO_OPTION_VBELINE_DRIVER_CIRRUS 10
#define VIDEO_OPTION_VBELINE_DRIVER_MATROX 11
#define VIDEO_OPTION_VBELINE_DRIVER_R128 12
#define VIDEO_OPTION_VBELINE_DRIVER_NEOMAGIC 13
#define VIDEO_OPTION_VBELINE_DRIVER_TRIDENT 14

void video_option_vbeline_driver_set(int mode);
int video_option_vbeline_driver_get(void);

/* Select the mode default */
#define VIDEO_OPTION_DEFAULT_MODE_SMART 0 /* Display only not duplicate modes (default) */
#define VIDEO_OPTION_DEFAULT_MODE_ALL 1 /* Display all modes */
void video_option_default_mode_set(int mode);
int video_option_default_mode_get(void);

/* Select a multiplier factor for the driver clock */
void video_option_vbeline_clock_multiplier_set(double factor);
double video_option_vbeline_clock_multiplier_get(void);
void video_option_vbeline_8bit_clock_multiplier_set(double factor);
double video_option_vbeline_8bit_clock_multiplier_get(void);
void video_option_vbeline_16bit_clock_multiplier_set(double factor);
double video_option_vbeline_16bit_clock_multiplier_get(void);

/***************************************************************************/
/* Video mode container */

typedef struct video_mode_container_struct {
	video_mode* base;
} video_mode_container;

void video_mode_container_init(video_mode_container* vmc);
void video_mode_container_done(video_mode_container* vmc);
video_error video_mode_container_load(video_mode_container* vmc, const char* (*get_string)(const char *section, const char *name, const char *def));
void video_mode_container_save(video_mode_container* vmc, void (*set_string)(const char *section, const char *name, const char *val));
const video_mode* video_mode_container_has(video_mode_container* vmc, const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b));
int video_mode_compare(const video_mode* vm0, const video_mode* vm1);
void video_mode_container_remove(video_mode_container* vmc, int (*modeselect)(const video_mode*, void*), void*);
void video_mode_container_insert(video_mode_container* vmc, const video_mode* vm);
void video_mode_container_insert_sort(video_mode_container* vmc, const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b));

typedef struct video_mode_container_iterator_struct {
	video_mode* base;
} video_mode_container_iterator;

void video_mode_container_iterator_begin(video_mode_container_iterator* vmci, video_mode_container* vmc);
void video_mode_container_iterator_next(video_mode_container_iterator* vmci);
int video_mode_container_iterator_end(video_mode_container_iterator* vmci);
video_mode* video_mode_container_iterator_get(video_mode_container_iterator* vmci);

video_error video_mode_parse_line(video_mode* mode, const char* begin);

/***************************************************************************/
/* Import */

video_error video_mode_import(video_mode* vm);
video_error video_mode_import_and_insert(video_mode_container* vmc, video_mode* vm);

/***************************************************************************/
/* Update */

unsigned video_update_x(void);
unsigned video_update_y(void);
unsigned video_update_page_max(void);
unsigned video_update_page(void);
void video_update_init(int max_buffer);
void video_update_done(void);
void video_update_start(void);
void video_update_stop(int wait_retrace);

/***************************************************************************/
/* Monitor */

typedef struct video_monitor_range_struct {
	double low;
	double high;
} video_monitor_range;

#define VIDEO_MONITOR_RANGE_MAX 8

typedef struct video_monitor_struct {
	video_monitor_range hclock[VIDEO_MONITOR_RANGE_MAX];
	video_monitor_range vclock[VIDEO_MONITOR_RANGE_MAX];
} video_monitor;

int video_monitor_is_empty(const video_monitor* monitor);

video_error video_monitor_load(video_monitor* monitor, const char* (*get_string)(const char *section, const char *name, const char *def));
video_error video_monitor_parse(video_monitor* monitor, const char* h, const char* v);

int video_monitor_hclock_check(const video_monitor* monitor, double hclock);
int video_monitor_vclock_check(const video_monitor* monitor, double vclock);
int video_monitor_clock_check(const video_monitor* monitor, double hclock, double vclock);
int video_monitor_crtc_check(const video_monitor* monitor, const video_crtc* crtc);
int video_monitor_mode_check(const video_monitor* monitor, const video_mode* mode);
int video_monitor_mode_check_safe(const video_monitor* monitor, const video_mode* mode);
int video_monitor_mode_adjust_check(const video_monitor* monitor, const video_mode* mode, double vert_freq);

double video_monitor_hclock_min(const video_monitor* monitor);
double video_monitor_hclock_max(const video_monitor* monitor);
double video_monitor_vclock_min(const video_monitor* monitor);
double video_monitor_vclock_max(const video_monitor* monitor);

/***************************************************************************/
/* Active video mode */

typedef unsigned video_rgb;

typedef struct video_internal_struct {
	int active; /* !=0 if active */

	int mode_active; /* !=0 if mode selected */
	video_mode mode_info; /* mode selected */

	int old_mode_grabbed; /* !=0 if old mode grabbed with succes */
	video_mode old_mode_info; /* startup mode */

	/* RGB */
	unsigned rgb_red_mask;
	unsigned rgb_green_mask;
	unsigned rgb_blue_mask;
	int rgb_red_shift;
	int rgb_green_shift;
	int rgb_blue_shift;
	unsigned rgb_red_len;
	unsigned rgb_green_len;
	unsigned rgb_blue_len;
	video_rgb rgb_mask_bit;
	video_rgb rgb_high_bit;
	video_rgb rgb_low_bit;

	char error[256]; /* last error */
} video_internal;

extern video_internal video_state;

/* Video active */
static __inline__ int video_active(void) {
	return video_state.active != 0;
}

/* Video mode active */
static __inline__ int video_mode_active(void) {
	assert( video_active() );
	return video_state.mode_active;
}

/* Current video mode */
static __inline__ const video_mode* video_current_mode(void) {
	assert( video_mode_active() );
	return &video_state.mode_info;
}

/* Driver */
static __inline__ unsigned video_driver(void) {
	return video_mode_driver(video_current_mode());
}

static __inline__ unsigned video_index(void) {
	return video_mode_index(video_current_mode());
}

/* Color depth in bits of the current video mode */
static __inline__ unsigned video_bits_per_pixel(void) {
	return video_mode_bits_per_pixel(video_current_mode());
}

/* Color depth in bytes of the current video mode */
static __inline__ unsigned video_bytes_per_pixel(void) {
	return video_mode_bytes_per_pixel(video_current_mode());
}

/* Type of mode */
static __inline__ int video_is_text(void) {
	return video_mode_is_text(video_current_mode());
}

static __inline__ int video_is_graphics(void) {
	return video_mode_is_graphics(video_current_mode());
}

/* Name of the video mode */
static __inline__  const char* video_name(void) {
	return video_mode_name(video_current_mode());
}

/* Frequencies */
static __inline__ double video_vert_freq(void) {
	return video_mode_vert_freq(video_current_mode());
}

static __inline__ double video_horz_freq(void) {
	return video_mode_horz_freq(video_current_mode());
}

/* Size of the screen */
static __inline__ unsigned video_size_x(void) {
	return video_mode_size_x(video_current_mode());
}

static __inline__ unsigned video_size_y(void) {
	return video_mode_size_y(video_current_mode());
}

/* Size of the font for text mode */
static __inline__ unsigned video_font_size_x(void) {
	if (vga_mode_active())
		return vga_font_size_x();
	else
		return vbe_font_size_x();
}

static __inline__ unsigned video_font_size_y(void) {
	if (vga_mode_active())
		return vga_font_size_y();
	else
		return vbe_font_size_y();
}

/* Virtual size of the screen */
static __inline__ unsigned video_virtual_x(void) {
	return video_mode_virtual_x(video_current_mode());
}

static __inline__ unsigned video_virtual_y(void) {
	return video_mode_virtual_y(video_current_mode());
}

/* Flags */
static __inline__ unsigned video_flags(void) {
	return video_mode_flags(video_current_mode());
}

/* Selector of the video memory */
extern unsigned FAST (*video_write_selector)(void);

/* Address for writing a line of the video memory */
extern unsigned FAST (*video_write_line)(unsigned y);

static __inline__ int video_unchained_active(void) {
	return vga_mode_active() && vga_unchained_active();
}

static __inline__ unsigned video_offset(unsigned x) {
	unsigned off = x * video_bytes_per_pixel();
	if (video_unchained_active())
		off /= 4;
	return off;
}

static __inline__ void video_unchained_plane_mask_set(unsigned plane_mask) {
	assert( vga_mode_active() );
	vga_unchained_plane_mask_set(plane_mask);
}

static __inline__ void video_unchained_plane_set(unsigned plane) {
	video_unchained_plane_mask_set(1 << plane);
}

/* Size of row */
static __inline__ unsigned video_bytes_per_scanline(void) {
	if (vga_mode_active())
		return vga_bytes_per_scanline();
	else
		return vbe_bytes_per_scanline();
}

unsigned video_bytes_per_page(void);

int video_index_packed_to_rgb_avaliable(void);
void video_index_packed_to_rgb(int waitvsync);

int video_index_rgb_to_packed_avaliable(void);
void video_index_rgb_to_packed(void);

void video_log_modeline_mb(const char *text, unsigned mode, unsigned pixel_clock, unsigned hde, unsigned hbs, unsigned hrs, unsigned hre, unsigned hbe, unsigned ht, unsigned vde, unsigned vbs, unsigned vrs, unsigned vre, unsigned vbe, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace);
void video_log_modeline_m(const char *text, unsigned mode, unsigned pixel_clock, unsigned hde, unsigned hrs, unsigned hre, unsigned ht, unsigned vde, unsigned vrs, unsigned vre, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace);
void video_log_modeline_b(const char *text, unsigned pixel_clock, unsigned hde, unsigned hbs, unsigned hrs, unsigned hre, unsigned hbe, unsigned ht, unsigned vde, unsigned vbs, unsigned vrs, unsigned vre, unsigned vbe, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace);
int video_print_modeline_mc(char* dest, unsigned mode, const video_crtc* crtc);

static __inline__ const char* video_error_description_get(void) {
	return video_state.error;
}

/* RGB nibble definition */
struct video_rgb_def_bits {
	unsigned red_len : 4; /* bit used for the red channel */
	unsigned red_pos : 5; /* shift for the red channel */
	unsigned green_len : 4;
	unsigned green_pos : 5;
	unsigned blue_len : 4;
	unsigned blue_pos : 5;
	/* 5*3+4*3 = 27 bit */
	unsigned dummy : 5;
};

/* Ordinal RGB nibble definiton */
typedef unsigned video_rgb_def;

union video_rgb_def_union {
	video_rgb_def ordinal;
	struct video_rgb_def_bits nibble;
};

video_rgb_def video_current_rgb_def_get(void);
const char* video_rgb_def_name_get(video_rgb_def rgb_def);
video_rgb_def video_rgb_def_get(unsigned red_len, unsigned red_pos, unsigned green_len, unsigned green_pos, unsigned blue_len, unsigned blue_pos);
video_rgb video_rgb_get_from_def(unsigned r, unsigned g, unsigned b, video_rgb_def def);

static __inline__ unsigned video_rgb_shift_get_from_def(unsigned len, unsigned pos) {
	return pos + len - 8;
}

static __inline__ unsigned video_rgb_mask_get_from_def(unsigned len, unsigned pos) {
	return ((1 << len) - 1) << pos;
}

video_color* video_palette_get(void);
video_error video_palette_set(video_color* palette, unsigned start, unsigned count, int waitvsync);

static __inline__ void video_palette_make(video_color* vp, unsigned r, unsigned g, unsigned b) {
	vp->red = r;
	vp->green = g;
	vp->blue = b;
}

unsigned video_color_dist(const video_color* A, const video_color* B);

static __inline__ unsigned video_rgb_shift(unsigned value, int shift) {
	if (shift >= 0)
		return value >> shift;
	else
		return value << -shift;
}

static __inline__ unsigned video_rgb_nibble_insert(unsigned value, int shift, unsigned mask) {
	return video_rgb_shift(value,-shift) & mask;
}

static __inline__ unsigned video_rgb_nibble_extract(unsigned value, int shift, unsigned mask) {
	return video_rgb_shift(value & mask,shift);
}

static __inline__ video_rgb video_rgb_get(unsigned r, unsigned g, unsigned b) {
	return video_rgb_nibble_insert(r, video_state.rgb_red_shift, video_state.rgb_red_mask) |
		video_rgb_nibble_insert(g, video_state.rgb_green_shift, video_state.rgb_green_mask) |
		video_rgb_nibble_insert(b, video_state.rgb_blue_shift, video_state.rgb_blue_mask);
}

static __inline__ void video_rgb_make(video_rgb* rgb, unsigned r, unsigned g, unsigned b) {
	*rgb = video_rgb_get(r,g,b);
}

static __inline__ video_rgb video_rgb_mask_bit_get(void) {
	return video_state.rgb_mask_bit;
}

static __inline__ video_rgb video_rgb_high_bit_get(void) {
	return video_state.rgb_high_bit;
}

static __inline__ video_rgb video_rgb_low_bit_get(void) {
	return video_state.rgb_low_bit;
}

static __inline__ unsigned video_red_get(unsigned rgb) {
	return video_rgb_nibble_extract(rgb, video_state.rgb_red_shift, video_state.rgb_red_mask);
}

static __inline__ unsigned video_green_get(unsigned rgb) {
	return video_rgb_nibble_extract(rgb, video_state.rgb_green_shift, video_state.rgb_green_mask);
}

static __inline__ unsigned video_blue_get(unsigned rgb) {
	return video_rgb_nibble_extract(rgb, video_state.rgb_blue_shift, video_state.rgb_blue_mask);
}

unsigned video_red_get_approx(unsigned rgb);
unsigned video_green_get_approx(unsigned rgb);
unsigned video_blue_get_approx(unsigned rgb);

void video_init(void);
void video_done(void);
video_error video_mode_set(video_mode* mode);
int video_mode_vert_freq_adjust_avaliable(video_mode* mode);
void video_mode_vert_freq_adjust(video_mode* mode, double required_vert_freq);
int video_mode_grab(video_mode* mode);
void video_mode_reset(void);
void video_mode_done(void);
void video_mode_print(char* buffer, const video_mode* vm);
void video_wait_vsync(void);
double video_measure_vert_freq(void);
double video_measure_vert_freq_safe(void);
video_error video_display_set_async(unsigned offset, int waitvsync);
video_error video_display_set_sync(unsigned offset);
void video_put_pixel(unsigned x, unsigned y, unsigned color);
void video_put_pixel_clip(unsigned x, unsigned y, unsigned color);
void video_put_char(unsigned x, unsigned y, char c, unsigned color);

video_error video_snapshot_save(const char* snapshot_name, int start_x, int start_y);

#ifdef __cplusplus
}
#endif

#endif
