/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"
#include "generate.h"
#include "vga.h"
#include "vgaline.h"
#include "vgatweak.h"
#include "vbe.h"
#include "vbeline.h"

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/movedata.h>
#include <sys/farptr.h>

/***************************************************************************/
/* Parse */

/* Line in parsing */
static char video_mode_parse_error[1024];

static video_error parse_int(int* r, const char** begin, const char* end) {
	char* e;
	*r = strtol(*begin,&e,10);
	if (e==*begin || e > end) {
		sprintf(video_mode_parse_error,"Error in value");
		return -1;
	}
	*begin = e;
	return 0;
}

static void parse_separator(const char* separator, const char** begin, const char* end) {
	while (*begin < end && strchr(separator,**begin))
		++*begin;
}

static video_error parse_token(char* token, unsigned size, const char* separator, const char** begin, const char* end) {
	unsigned pos = 0;
	while (*begin < end && !strchr(separator,**begin)) {
		if (pos + 2 > size) {
			sprintf(video_mode_parse_error,"Token too long");
			return -1;
		}
		token[pos] = **begin;
		++pos;
		++*begin;
	}
	if (!pos) {
		sprintf(video_mode_parse_error,"Expected token");
		return -1;
	}
	token[pos] = 0;
	return 0;
}

static video_error parse_hex(unsigned* r, const char** begin, const char* end) {
	char* e;
	*r = strtoul(*begin,&e,16);
	if (e==*begin || e > end) {
		sprintf(video_mode_parse_error,"Error in value");
		return -1;
	}
	*begin = e;
	return 0;
}

static video_error parse_double(double* r, const char** begin, const char* end) {
	char* e;
	*r = strtod(*begin,&e);
	if (e==*begin || e > end) {
		sprintf(video_mode_parse_error,"Error in value");
		return -1;
	}
	*begin = e;
	return 0;
}

/***************************************************************************/
/* Video mode container */

/* Inizialize the video mode container */
void video_mode_container_init(video_mode_container* vmc) {
	vmc->base = 0;
}

/* Deinizialize the video mode container */
void video_mode_container_done(video_mode_container* vmc) {
	video_mode* p = vmc->base;
	while (p) {
		video_mode* todelete = p;
		p = p->next;
		free(todelete);
	}
}

/* Check if mode is already in container
	Don't compare name, but the mode parameter
	Return ==0 if not found, otherwise the mode in the container
*/
const video_mode* video_mode_container_has(video_mode_container* vmc, const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b)) {
	video_mode* p = vmc->base;
	while (p) {
		if (compare(p,vm)==0)
			return p;
		p = p->next;
	}
	return 0;
}

/* Remove a mode in the container
	Return the new container
*/
void video_mode_container_remove(video_mode_container* vmc, int (*select)(const video_mode*, void*), void* arg) {
	while (vmc->base && select(vmc->base,arg)) {
		video_mode* todelete = vmc->base;
		vmc->base = vmc->base->next;
		free(todelete);
	}
	if (vmc->base) {
		video_mode* p = vmc->base;
		while (p->next) {
			if (select(p->next,arg)) {
				video_mode* todelete = p->next;
				p->next = p->next->next;
				free(todelete);
			} else
				p = p->next;
		}
	}
}

/* Insert a mode in the container */
void video_mode_container_insert(video_mode_container* vmc, const video_mode* vm) {
	video_mode* p = malloc(sizeof(video_mode));
	*p = *vm;
	p->next = vmc->base;
	vmc->base = p;
}

/* Insert in order */
void video_mode_container_insert_sort(video_mode_container* vmc,  const video_mode* vm, int (*compare)(const video_mode* a,const video_mode* b)) {
	if (vmc->base && compare(vmc->base,vm)<=0) {
		video_mode* p = vmc->base;
		video_mode* c;
		while (p->next && compare(p->next,vm)<=0) {
			p = p->next;
		}
		c = malloc(sizeof(video_mode));
		*c = *vm;
		c->next = p->next;
		p->next = c;
	} else {
		video_mode* p = malloc(sizeof(video_mode));
		*p = *vm;
		p->next = vmc->base;
		vmc->base = p;
	}
}

/* Compare two video mode
	Return ==0 if equal
*/
int video_mode_compare(const video_mode* vm0, const video_mode* vm1) {
	int m0 = vm0->flags & VIDEO_FLAGS_DRIVER_MASK;
	int m1 = vm1->flags & VIDEO_FLAGS_DRIVER_MASK;
	if (m0!=m1)
		return 1;
	switch (m0) {
		case VIDEO_FLAGS_DRIVER_VGA :
			return memcmp(&vm0->driver.vga,&vm1->driver.vga,sizeof(vga_video_mode))!=0;
		case VIDEO_FLAGS_DRIVER_VGATWEAK : {
			/* TODO do an effective compare ? order may be important ! */
			unsigned i;
			if (vm0->driver.vgatweak.mode != vm1->driver.vgatweak.mode)
				return 1;
			if (vm0->driver.vgatweak.regs_num != vm1->driver.vgatweak.regs_num)
				return 1;
			for(i=0;i<vm0->driver.vgatweak.regs_num;++i) {
				if (vm0->driver.vgatweak.regs[i].port != vm1->driver.vgatweak.regs[i].port)
					return 1;
				if (vm0->driver.vgatweak.regs[i].index != vm1->driver.vgatweak.regs[i].index)
					return 1;
				if (vm0->driver.vgatweak.regs[i].value != vm1->driver.vgatweak.regs[i].value)
					return 1;
			}
			return 0;
		}
		case VIDEO_FLAGS_DRIVER_VGALINE :
			return memcmp(&vm0->driver.vgaline,&vm1->driver.vgaline,sizeof(vgaline_video_mode))!=0;
		case VIDEO_FLAGS_DRIVER_VBE :
			return memcmp(&vm0->driver.vbe,&vm1->driver.vbe,sizeof(vbe_video_mode))!=0;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return memcmp(&vm0->driver.vbeline,&vm1->driver.vbeline,sizeof(vbeline_video_mode))!=0;
	}
	return 1;
}

static video_error parse_reg(tweak_register* r, const char** begin, const char* end) {
	unsigned v;
	parse_separator(" \t",begin,end);
	if (parse_hex(&v,begin,end))
		return -1;
	r->port = v;
	parse_separator(" \t",begin,end);
	if (**begin!='[') {
		sprintf(video_mode_parse_error,"Miss [");
		return -1;
	}
	++*begin;
	parse_separator(" \t",begin,end);
	if (parse_hex(&v,begin,end))
		return -1;
	r->index = v;
	parse_separator(" \t",begin,end);
	if (**begin!=']') {
		sprintf(video_mode_parse_error,"Miss ]");
		return -1;
	}
	++*begin;
	parse_separator(" \t",begin,end);
	if (**begin!='=') {
		sprintf(video_mode_parse_error,"Miss =");
		return -1;
	}
	++*begin;
	parse_separator(" \t",begin,end);
	if (parse_hex(&v,begin,end))
		return -1;
	r->value = v;
	parse_separator(" \t",begin,end);
	return 0;
}

static video_error parse_vgatweak(vgatweak_video_mode* mode, const char* begin, const char* end) {
	unsigned v;
	parse_separator(" \t",&begin,end);
	if (parse_hex(&v,&begin,end))
		return -1;
	mode->mode = v;
	mode->regs_num = 0;
	while (begin != end) {
		if (mode->regs_num >= VIDEO_MODE_REGS_MAX) {
			sprintf(video_mode_parse_error,"Too many regs");
			return -1;
		}
		if (parse_reg( mode->regs + mode->regs_num, &begin, end))
			return -1;
		++mode->regs_num;
	}
	return 0;
}

static video_error parse_modecrtc(unsigned*mode, video_crtc* crtc, const char* begin, const char* end) {
	int v;
	unsigned u;
	double d;
	parse_separator(" \t",&begin,end);
	if (parse_hex(&u,&begin,end))
		return -1;
	*mode = u;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	crtc->pixelclock = d * 1E6;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->hde = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->hrs = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->hre = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->ht = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vde = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vrs = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vre = v;
	parse_separator(" \t",&begin,end);
	if (parse_int(&v,&begin,end))
		return -1;
	crtc->vt = v;

	/* defaults */
	crtc->doublescan = 0;
	crtc->interlaced = 0;
	crtc->hsync = 1;
	crtc->vsync = 0;

	parse_separator(" \t",&begin,end);
	while (begin != end) {
		char t[16];
		if (parse_token(t,sizeof(t)," \t",&begin,end))
			return -1;
		if (stricmp(t,"doublescan")==0) {
			crtc->doublescan = 1;
		} else if (stricmp(t,"+hsync")==0) {
			crtc->hsync = 0;
		} else if (stricmp(t,"-hsync")==0) {
			crtc->hsync = 1;
		} else if (stricmp(t,"+vsync")==0) {
			crtc->vsync = 0;
		} else if (stricmp(t,"-vsync")==0) {
			crtc->vsync = 1;
		} else if (stricmp(t,"interlaced")==0) {
			crtc->interlaced = 1;
		} else {
			sprintf(video_mode_parse_error,"Unknow token");
			return -1;
		}
		parse_separator(" \t",&begin,end);
	}
	return 0;
}

/* Parse a line */
video_error video_mode_parse_line(video_mode* mode, const char* begin) {
	char name[VIDEO_MODE_NAME_MAX];
	char type[64];
	const char* end = begin + strlen(begin);
	const char* argend;

	parse_separator(" \t",&begin,end);
	if (parse_token(name,sizeof(name)," \t",&begin,end))
		return -1;
	parse_separator(" \t",&begin,end);
	if (parse_token(type,sizeof(type)," \t(",&begin,end))
		return -1;
	parse_separator(" \t",&begin,end);
	if (*begin!='(') {
		sprintf(video_mode_parse_error,"Miss (");
		return -1;
	}
	++begin;

	argend = strchr(begin,')');
	if (!argend || argend >= end) {
		sprintf(video_mode_parse_error,"Miss )");
		return -1;
	}

	memset(mode,0,sizeof(video_mode));
	if (strlen(name) >= VIDEO_MODE_NAME_MAX) {
		sprintf(video_mode_parse_error,"Video mode name too long");
		return -1;
	}
	strcpy(mode->name,name);

	/* type */
	if (strcmp(type,"vbe")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VBE;
		if (parse_hex(&mode->driver.vbe.mode,&begin,argend))
			return -1;
	} else if (strcmp(type,"vga")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA;
		if (parse_hex(&mode->driver.vga.mode,&begin,argend))
			return -1;
	} else if (strcmp(type,"vgatweak")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGATWEAK;
		if (parse_vgatweak(&mode->driver.vgatweak,begin,argend))
			return -1;
	} else if (strcmp(type,"vgaline")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGALINE;
		if (parse_modecrtc(&mode->driver.vgaline.mode,&mode->driver.vgaline.crtc,begin,argend))
			return -1;
	} else if (strcmp(type,"vbeline")==0) {
		mode->flags = VIDEO_FLAGS_DRIVER_VBELINE;
		if (parse_modecrtc(&mode->driver.vbeline.mode,&mode->driver.vbeline.crtc,begin,argend))
			return -1;
	} else {
		sprintf(video_mode_parse_error,"Unknow mode type");
		return -1;
	}

	return 0;
}

/* Max number of mode saved in config file */
#define VIDEO_CONFIG_MAX 256

/* Load the list of video mode */
video_error video_mode_container_load(video_mode_container* vmc, const char* (*get_string)(const char *section, const char *name, const char *def)) {
	unsigned i;
	int error = 0;
	for(i=0;i<VIDEO_CONFIG_MAX;++i) {
		char modeline[32];
		const char* s;
		sprintf(modeline,"modeline%d",i);
		s = get_string("video", modeline, 0);
		if (s) {
			video_mode mode;
			char* buffer = strdup(s);
			memset(&mode,0,sizeof(mode));
			if (video_mode_parse_line(&mode,buffer)) {
				if (!error) {
					error = 1;
					video_error_description_set("%s parsing modeline %s",video_mode_parse_error,s);
				}
			} else {
				if (video_mode_import_and_insert(vmc,&mode)!=0) {
					if (!error) {
						error = 1;
						video_error_description_set("Unsupported video mode %s",mode.name);
					}
				}
			}
			free(buffer);
		}
	}
	return error ? -1 : 0;
}

/* Export the modeline */
void video_mode_print(char* buffer, const video_mode* vm) {
	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA:
			video_mode_vga_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE:
			video_mode_vgaline_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VGATWEAK:
			video_mode_vgatweak_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
			video_mode_vbe_print(buffer,vm);
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			video_mode_vbeline_print(buffer,vm);
		break;
	}
}

/* Save the list of video mode */
void video_mode_container_save(video_mode_container* vmc, void (*set_string)(const char *section, const char *name, const char *val)) {
	video_mode_container_iterator j;
	int i;
	video_mode_container_iterator_begin(&j,vmc);
	for(i=0;i<VIDEO_CONFIG_MAX;++i) {
		char modeline[32];
		sprintf(modeline,"modeline%d",i);
		if (!video_mode_container_iterator_end(&j)) {
			char buffer[1024];
			video_mode* vm = video_mode_container_iterator_get(&j);
			switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
				case VIDEO_FLAGS_DRIVER_VGA:
					video_mode_vga_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VGALINE:
					video_mode_vgaline_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VGATWEAK:
					video_mode_vgatweak_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VBE :
					video_mode_vbe_print(buffer,vm);
				break;
				case VIDEO_FLAGS_DRIVER_VBELINE :
					video_mode_vbeline_print(buffer,vm);
				break;
			}
			set_string("video",modeline,buffer);
			video_mode_container_iterator_next(&j);
		} else {
			set_string("video",modeline,0);
		}
	}
}

void video_mode_container_iterator_begin(video_mode_container_iterator* vmci, video_mode_container* vmc) {
	vmci->base = vmc->base;
}

void video_mode_container_iterator_next(video_mode_container_iterator* vmci) {
	vmci->base = vmci->base->next;
}

int video_mode_container_iterator_end(video_mode_container_iterator* vmci) {
	return vmci->base == 0;
}

video_mode* video_mode_container_iterator_get(video_mode_container_iterator* vmci) {
	return vmci->base;
}

/***************************************************************************/
/* Options */

/* Load the video options configuration
 * return:
 *   ==0 success
 */
video_error video_options_load(const char* (*get_string)(const char *section, const char *name, const char *def)) {
	const char* s;

	/* Options must be loaded before initialization */
	assert( !video_active() );

	s = get_string("video","vbeline_tweak","aggressive");
	if (s) {
		if (stricmp(s,"aggressive")==0) {
			video_option_vbeline_tweak_set(VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE);
		} else if (stricmp(s,"moderate")==0) {
			video_option_vbeline_tweak_set(VIDEO_OPTION_VBELINE_TWEAK_MODERATE);
		} else {
			video_error_description_set("Invalid option 'vbeline_tweak = %s'",s);
			return -1;
		}
	}

	s = get_string("video","vbeline_driver","auto");
	if (s) {
		if (stricmp(s,"no")==0 || stricmp(s,"none")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_NONE);
		} else if (stricmp(s,"auto")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_AUTO);
		} else if (stricmp(s,"vbe3")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_VBE3);
		} else if (stricmp(s,"laguna")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_LAGUNA);
		} else if (stricmp(s,"sis6326")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_SIS6326);
		} else if (stricmp(s,"savage")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_SAVAGE);
		} else if (stricmp(s,"3dfx")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_3DFX);
		} else if (stricmp(s,"ati")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_ATI);
		} else if (stricmp(s,"s3")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_S3);
		} else if (stricmp(s,"cirrus")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_CIRRUS);
		} else if (stricmp(s,"matrox")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_MATROX);
		} else if (stricmp(s,"r128")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_R128);
		} else if (stricmp(s,"neomagic")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_NEOMAGIC);
		} else if (stricmp(s,"trident")==0) {
			video_option_vbeline_driver_set(VIDEO_OPTION_VBELINE_DRIVER_TRIDENT);
		} else {
			video_error_description_set("Invalid option 'vbeline_driver = %s'",s);
			return -1;
		}
	}

	s = get_string("video","default_mode","smart");
	if (s) {
		if (stricmp(s,"smart")==0) {
			video_option_default_mode_set(VIDEO_OPTION_DEFAULT_MODE_SMART);
		} else if (stricmp(s,"all")==0) {
			video_option_default_mode_set(VIDEO_OPTION_DEFAULT_MODE_ALL);
		} else {
			video_error_description_set("Invalid option 'mode_default = %s'",s);
			return -1;
		}
	}

	s = get_string("video","crtc_blank","aggressive");
	if (s) {
		if (stricmp(s,"aggressive")==0) {
			video_option_crtc_blank_set(VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE);
		} else if (stricmp(s,"moderate")==0) {
			video_option_crtc_blank_set(VIDEO_OPTION_CRTC_BLANK_MODERATE);
		} else {
			video_error_description_set("Invalid option 'crtc_blank = %s'",s);
			return -1;
		}
	}

	s = get_string("video","vbeline_clock_multiplier","1.0");
	if (s) {
		double v = strtod(s,0);
		if (v<=0) {
			video_error_description_set("Invalid option 'vbeline_clock_multiplier = %s'",s);
			return -1;
		}
		video_option_vbeline_clock_multiplier_set(v);
	}

	s = get_string("video","vbeline_8bit_clock_multiplier","1.0");
	if (s) {
		double v = strtod(s,0);
		if (v<=0) {
			video_error_description_set("Invalid option 'vbeline_8bit_clock_multiplier = %s'",s);
			return -1;
		}
		video_option_vbeline_8bit_clock_multiplier_set(v);
	}

	s = get_string("video","vbeline_16bit_clock_multiplier","1.0");
	if (s) {
		double v = strtod(s,0);
		if (v<=0) {
			video_error_description_set("Invalid option 'vbeline_16bit_clock_multiplier = %s'",s);
			return -1;
		}
		video_option_vbeline_16bit_clock_multiplier_set(v);
	}

	return 0;
}

/***************************************************************************/
/* Monitor */

static video_error parse_monitor(video_monitor_range* range, const char* begin, const char* end, double mult) {
	unsigned i = 0;
	parse_separator(" \t",&begin,end);
	while (begin < end) {
		double v0;
		if (i==VIDEO_MONITOR_RANGE_MAX) {
			return -1;
		}
		if (parse_double(&v0,&begin,end))
			return -1;
		parse_separator(" \t",&begin,end);
		if (*begin=='-') {
			double v1;
			++begin;
			if (parse_double(&v1,&begin,end))
				return -1;
			if (v0 < 0 || v0 > v1 || v1 > 300) {
				return -1;
			}
			range[i].low = mult * v0;
			range[i].high = mult * v1;
			++i;
		} else {
			if (v0 < 0 || v0 > 300) {
				return -1;
			}
			/* the 0 value is used as a null definition */
			if (v0 > 0) {
				range[i].low = mult * v0;
				range[i].high = range[i].low;
				++i;
			}
		}
		parse_separator(" \t",&begin,end);
		if (*begin==',') {
			++begin;
		}
		parse_separator(" \t",&begin,end);
	}
	if (i==0)
		return -1; // empty
	while (i<VIDEO_MONITOR_RANGE_MAX) {
		range[i].low = 0;
		range[i].high = 0;
		++i;
	}
	return 0;
}

int video_monitor_is_empty(const video_monitor* monitor) {
	int i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->hclock[i].low != 0 || monitor->hclock[i].high != 0)
			return 0;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->vclock[i].low != 0 || monitor->vclock[i].high != 0)
			return 0;
	return 1;
}

video_error video_monitor_parse(video_monitor* monitor, const char* h, const char* v) {
	if (parse_monitor(monitor->hclock,h,h+strlen(h),1000)!=0) {
		video_error_description_set("Invalid monitor 'hclock' specification");
		return -1;
	}
	if (parse_monitor(monitor->vclock,v,v+strlen(v),1)!=0) {
		video_error_description_set("Invalid monitor 'vclock' specification");
		return -1;
	}
	return 0;
}

/* Load the monitor configuration
 * return:
 *   ==0 success
 *   <0 error
 *   >0 configuration missing, video_monitor_is_empty(monitor)!=0
 */
video_error video_monitor_load(video_monitor* monitor, const char* (*get_string)(const char *section, const char *name, const char *def)) {
	const char* s;

	s = get_string("video","hclock",0);
	if (!s) {
		/* Fill all with 0, video_monitor_is_empty()==1 */
		memset(monitor,sizeof(video_monitor),0);
		video_error_description_set("Missing monitor '[video] hclock' specification");
		return 1;
	}
	if (parse_monitor(monitor->hclock,s,s+strlen(s),1000)!=0) {
		video_error_description_set("Invalid monitor '[video] hclock' specification");
		return -1;
	}
	s = get_string("video","vclock",0);
	if (!s) {
		/* Fill all with 0, video_monitor_is_empty()==1 */
		memset(monitor,sizeof(video_monitor),0);
		video_error_description_set("Missing monitor '[video] vclock' specification");
		return 1;
	}
	if (parse_monitor(monitor->vclock,s,s+strlen(s),1)!=0) {
		video_error_description_set("Invalid monitor '[video] vclock' specification");
		return -1;
	}
	return 0;
}

/* Check if a video mode is acceptable
 * return:
 *   !=0 acceptable
 *   ==0 no
 */
int video_monitor_hclock_check(const video_monitor* monitor, double hclock) {
	const double monitor_hfix_error = 0.02;
	const double monitor_hrange_error = 0.01;
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
		if (monitor->hclock[i].low != 0 && monitor->hclock[i].high != 0) {
			if (monitor->hclock[i].low == monitor->hclock[i].high) {
				// use a % error
				if (monitor->hclock[i].low * (1-monitor_hfix_error) <= hclock && hclock <= monitor->hclock[i].high * (1+monitor_hfix_error))
					break;
			} else {
				// use 1 unit error
				if (monitor->hclock[i].low * (1-monitor_hrange_error) <= hclock && hclock <= monitor->hclock[i].high * (1+monitor_hrange_error))
					break;
			}
		}
	}
	if (i==VIDEO_MONITOR_RANGE_MAX) {
		video_error_description_nolog_set("Horizontal clock of %.2f kHz is out of range of your monitor", (double)hclock / 1E3);
		return 0;
	}
	return 1;
}

int video_monitor_vclock_check(const video_monitor* monitor, double vclock) {
	const double monitor_vfix_error = 0.02;
	const double monitor_vrange_error = 0.01;
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i) {
		if (monitor->vclock[i].low != 0 && monitor->vclock[i].high != 0) {
			if (monitor->vclock[i].low == monitor->vclock[i].high) {
				if (monitor->vclock[i].low * (1-monitor_vfix_error) <= vclock && vclock <= monitor->vclock[i].high * (1+monitor_vfix_error))
					break;
			} else {
				if (monitor->vclock[i].low * (1-monitor_vrange_error)<= vclock && vclock <= monitor->vclock[i].high * (1+monitor_vrange_error))
					break;
			}
		}
	}
	if (i==VIDEO_MONITOR_RANGE_MAX) {
		video_error_description_nolog_set("Vertical clock of %.2f Hz is out of range of your monitor", (double)vclock);
		return 0;
	}
	return 1;
}

int video_monitor_clock_check(const video_monitor* monitor, double hclock, double vclock) {
	return video_monitor_hclock_check(monitor,hclock) && video_monitor_vclock_check(monitor,vclock);
}

/* Check if a video mode is acceptable
 * return:
 *   !=0 acceptable
 *   ==0 no
 */
int video_monitor_mode_check(const video_monitor* monitor, const video_mode* mode) {
	switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VBE :
			return 1;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			return video_monitor_clock_check(monitor,
				mode->horz_freq,
				mode->vert_freq);
		case VIDEO_FLAGS_DRIVER_VGATWEAK : {
			struct tweak_crtc info;
			tweak_crtc_get(&info, tweak_register_get, (void*)&mode->driver.vgatweak);
			return video_monitor_clock_check(monitor,info.horz_clock,info.vert_clock);
		}
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return video_monitor_clock_check(monitor,
				mode->horz_freq,
				mode->vert_freq);
	}
	return 0;
}

/* Check if a video mode is acceptable in a safe mode
 * return:
 *   !=0 acceptable
 *   ==0 no
 */
int video_monitor_mode_check_safe(const video_monitor* monitor, const video_mode* mode) {
	switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VBE :
			return video_monitor_hclock_check(monitor, 31500);
		case VIDEO_FLAGS_DRIVER_VGALINE :
			return video_monitor_clock_check(monitor,
				mode->horz_freq,
				mode->vert_freq);
		case VIDEO_FLAGS_DRIVER_VGATWEAK : {
			struct tweak_crtc info;
			tweak_crtc_get(&info, tweak_register_get, (void*)&mode->driver.vgatweak);
			return video_monitor_clock_check(monitor,info.horz_clock,info.vert_clock);
		}
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return video_monitor_clock_check(monitor,
				mode->horz_freq,
				mode->vert_freq);
	}
	return 0;
}

int video_monitor_mode_adjust_check(const video_monitor* monitor, const video_mode* mode, double vert_freq) {
	video_mode temp = *mode;
	video_mode_vert_freq_adjust(&temp,vert_freq);
	return video_monitor_mode_check(monitor,&temp);
}

double video_monitor_hclock_min(const video_monitor* monitor) {
	double min = 0;
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->hclock[i].low != 0 && (min == 0 || monitor->hclock[i].low < min))
			min = monitor->hclock[i].low;
	return min;
}

double video_monitor_hclock_max(const video_monitor* monitor) {
	double max = 0;
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->hclock[i].high != 0 && (max == 0 || monitor->hclock[i].high > max))
			max = monitor->hclock[i].high;
	return max;
}

double video_monitor_vclock_min(const video_monitor* monitor) {
	double min = 0;
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->vclock[i].low != 0 && (min == 0 || monitor->vclock[i].low < min))
			min = monitor->vclock[i].low;
	return min;
}

double video_monitor_vclock_max(const video_monitor* monitor) {
	double max = 0;
	unsigned i;
	for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
		if (monitor->vclock[i].high != 0 && (max == 0 || monitor->vclock[i].high > max))
			max = monitor->vclock[i].high;
	return max;
}

int video_monitor_crtc_check(const video_monitor* monitor, const video_crtc* crtc) {
	double hclock =	(double)crtc->pixelclock / crtc->ht;
	double vclock =	hclock / crtc->vt;
	if (crtc->doublescan)
		vclock /= 2;
	if (crtc->interlaced)
		vclock *= 2;
	return video_monitor_clock_check(monitor,hclock,vclock);
}

/***************************************************************************/
/* Generate */

static video_error parse_generate(const char* begin, const char* end, video_generate* data) {
	double d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->hactive = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->hfront = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->hsync = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->hback = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->vactive = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->vfront = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->vsync = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->vback = d;
	parse_separator(" \t",&begin,end);
	if (begin != end)
		return -1;

	/* compute the delta version */
	data->hback -= data->hsync;
	data->hsync -= data->hfront;
	data->hfront -= data->hactive;
	data->vback -= data->vsync;
	data->vsync -= data->vfront;
	data->vfront -= data->vactive;

	return 0;
}

video_error generate_parse(video_generate* generate, const char* g) {
	if (parse_generate(g,g+strlen(g),generate)!=0) {
		video_error_description_set("Invalid 'format' specification");
		return -1;
	}

	return 0;
}

video_error generate_load(video_generate* generate, const char* (*get_string)(const char *section, const char *name, const char *def)) {
	const char* s;

	s = get_string("video","format",0);
	if (!s) {
		video_error_description_set("Missing '[video] format' specification");
		return 1;
	}

	if (parse_generate(s,s+strlen(s),generate)!=0) {
		video_error_description_set("Invalid '[video] format' specification");
		return -1;
	}

	return 0;
}

static video_error parse_generate_linear(const char* begin, const char* end, video_generate_linear* data) {
	double d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.hactive = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.hfront = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.hsync = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.hback = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.vactive = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.vfront = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen0.vsync = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	parse_separator(" \t",&begin,end);
	data->gen0.vback = d;
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.hactive = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.hfront = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.hsync = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.hback = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.vactive = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.vfront = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.vsync = d;
	parse_separator(" \t",&begin,end);
	if (parse_double(&d,&begin,end))
		return -1;
	data->gen1.vback = d;
	parse_separator(" \t",&begin,end);
	if (begin != end)
		return -1;

	data->hsize0 = data->gen0.hactive;
	data->vsize0 = data->gen0.vactive;
	data->hsize1 = data->gen1.hactive;
	data->vsize1 = data->gen1.vactive;

	/* compute the delta version */
	data->gen0.hback -= data->gen0.hsync;
	data->gen0.hsync -= data->gen0.hfront;
	data->gen0.hfront -= data->gen0.hactive;
	data->gen0.vback -= data->gen0.vsync;
	data->gen0.vsync -= data->gen0.vfront;
	data->gen0.vfront -= data->gen0.vactive;

	data->gen1.hback -= data->gen1.hsync;
	data->gen1.hsync -= data->gen1.hfront;
	data->gen1.hfront -= data->gen1.hactive;
	data->gen1.vback -= data->gen1.vsync;
	data->gen1.vsync -= data->gen1.vfront;
	data->gen1.vfront -= data->gen1.vactive;

	return 0;
}

video_error generate_linear_load(video_generate_linear* generate_linear, const char* (*get_string)(const char *section, const char *name, const char *def)) {
	const char* s;

	s = get_string("video","format_linear",0);
	if (!s) {
		video_error_description_set("Missing '[video] format_linear' specification");
		return 1;
	}

	if (parse_generate_linear(s,s+strlen(s),generate_linear)!=0) {
		video_error_description_set("Invalid '[video] format_linear' specification");
		return -1;
	}

	return 0;
}

/***************************************************************************/
/* Snapshot */

typedef struct tagBITMAPFILEHEADER {
    uint16    bfType __attribute__ ((packed));
    uint32   bfSize __attribute__ ((packed));
    uint16    bfReserved1 __attribute__ ((packed));
    uint16    bfReserved2 __attribute__ ((packed));
    uint32   bfOffBits __attribute__ ((packed));
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER {
    uint32   biSize __attribute__ ((packed));
    uint32    biWidth __attribute__ ((packed));
    uint32    biHeight __attribute__ ((packed));
    uint16    biPlanes __attribute__ ((packed));
    uint16    biBitCount __attribute__ ((packed));
    uint32   biCompression __attribute__ ((packed));
    uint32   biSizeImage __attribute__ ((packed));
    uint32    biXPelsPerMeter __attribute__ ((packed));
    uint32    biYPelsPerMeter __attribute__ ((packed));
    uint32   biClrUsed __attribute__ ((packed));
    uint32   biClrImportant __attribute__ ((packed));
} BITMAPINFOHEADER;

video_error video_snapshot_save(const char* snapshot_name, int start_x, int start_y) {
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	FILE* f;
	unsigned line_size;
	int y;

	if (video_unchained_active())
		return -1; /* unchained unsupported */

	f = fopen(snapshot_name,"wb");
	if (!f)
		return -1;

	line_size = video_size_x() * 3;
	line_size = (line_size + 0x3) & ~0x3;

	memset(&bmfh,0,sizeof(bmfh));
	bmfh.bfType = 19778;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + line_size * video_size_y();
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

	memset(&bmih,0,sizeof(bmih));
	bmih.biSize = sizeof(bmih);
	bmih.biWidth = video_size_x();
	bmih.biHeight = video_size_y();
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;

	fwrite(&bmfh,sizeof(bmfh),1,f);
	fwrite(&bmih,sizeof(bmih),1,f);

	for(y=video_size_y()-1;y>=0;--y) {
		unsigned count = 0;
		unsigned long line = video_write_line(y + start_y);
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
			unsigned x;
			for(x=0;x<video_size_x();++x) {
				uint8 color = _farpeekb(video_write_selector(),line + x + start_x);
				video_color* palette = video_palette_get() + _farpeekb(video_write_selector(),line + color);
				uint8 b;
				b = palette->blue;
				fwrite(&b,1,1,f);
				b = palette->green;
				fwrite(&b,1,1,f);
				b = palette->red;
				fwrite(&b,1,1,f);
				count += 3;
			}
		} else {
			if (video_bytes_per_pixel() == 1) {
				unsigned x;
				for(x=0;x<video_size_x();++x) {
					uint8 color = _farpeekb(video_write_selector(),line + x + start_x);
					uint8 b;
					b = video_blue_get(color);
					fwrite(&b,1,1,f);
					b = video_green_get(color);
					fwrite(&b,1,1,f);
					b = video_red_get(color);
					fwrite(&b,1,1,f);
					count += 3;
				}
			} else if (video_bytes_per_pixel() == 2) {
				unsigned x;
				for(x=0;x<video_size_x();++x) {
					uint16 color = _farpeekw(video_write_selector(),line + (x+start_x)*2);
					uint8 b;
					b = video_blue_get(color);
					fwrite(&b,1,1,f);
					b = video_green_get(color);
					fwrite(&b,1,1,f);
					b = video_red_get(color);
					fwrite(&b,1,1,f);
					count += 3;
				}
			}
		}
		while (count % 4) {
			uint8 b = 0;
			fwrite(&b,1,1,f);
			++count;
		}
	}

	fclose(f);

	return 0;
}

