/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrvbe.h"

#include <stdlib.h>
#include <dpmi.h>
#include <go32.h>
#include <dos.h>
#include <pc.h>
#include <sys/movedata.h>
#include <sys/farptr.h>

/* If defined correct some bogus bios */
#define VBE_BOGUS

/* VBE Internal Mode Information */
vbe_internal vbe_state;

/* Video mode list */
#define VBE_MODE_MAX 512
unsigned vbe_mode_map[VBE_MODE_MAX];

unsigned FAST (*vbe_write_selector)(void);

/* Return the offset for accessing in writing the video memory */
unsigned FAST (*vbe_write_offset)(unsigned offset, unsigned size);

/* Return the offset for accessing in writing the video memory */
unsigned FAST (*vbe_write_line)(unsigned y);

/***************************************************************************/
/* Physical address mapping */

static int physical_init(unsigned* rselector, unsigned address, unsigned size) {
	__dpmi_meminfo meminfo;
	int selector;
	unsigned offset;

	assert( address >= 0x100000 );

	meminfo.address = address;
	meminfo.size = size;

	if (__dpmi_physical_address_mapping(&meminfo) != 0) {
		video_error_description_set("Error in dpmi_physical_address_mapping");
		return -1;
	}

	offset = meminfo.address;

	/* lock the linear memory range */
	__dpmi_lock_linear_region(&meminfo);

	/* allocate an ldt descriptor */
	selector = __dpmi_allocate_ldt_descriptors(1);
	if (selector == -1) {
		video_error_description_set("Error in dpmi_allocate_ldt_descriptors");
		__dpmi_free_physical_address_mapping(&meminfo);
		return -1;
	}

	/* set the descriptor base and limit */
	__dpmi_set_segment_base_address(selector, offset);

	/* prevent selector to change to no big (if size <= 64k) */
	if (size < 64*1024)
		__dpmi_set_segment_limit(selector, 64*1024 - 1);
	else
		__dpmi_set_segment_limit(selector, size - 1);

	*rselector = selector;

	return 0;
}

static void physical_done(unsigned selector, unsigned address, unsigned size) {
	__dpmi_meminfo meminfo;

	meminfo.address = address;
	meminfo.size = size;

	/* unlock */
	__dpmi_unlock_linear_region(&meminfo);

	/* free */
	__dpmi_free_physical_address_mapping(&meminfo);

	/* free selector */
	__dpmi_free_ldt_descriptor(selector);
}

/***************************************************************************/
/* Scanline */

/* Adjust scanline size */
static unsigned vbe_default_adjust_scanline_size(unsigned size) {
#ifdef VBE_BOGUS
	/* align at 16 bytes */
	return (size + 0xF) & ~0xF;
#else
	/* align at 4 bytes */
	return (size + 3) & ~0x3;
#endif
}

/* Adjust scanline size for banked mode */
static unsigned vbe_bank_adjust_scanline_size(unsigned size) {
	unsigned max = vbe_state.mode_info.WinSize * 1024;
	
	assert( vbe_mode_active() );
	
	while (size < max && max % size != 0)
		size = vbe_default_adjust_scanline_size(size + vbe_state.bytes_per_pixel);
	if (size < max)
		return size;
	return 0;
}

static unsigned vbe_adjust_scanline_size(unsigned size) {
	assert( vbe_mode_active() );
	if (vbe_bank_active()) {
		return vbe_bank_adjust_scanline_size(size);
	} else {
		return vbe_default_adjust_scanline_size(size);
	}
}

/* Get scanlines information
 * out:
 *   *byte_length length in byte
 *   *pixel_length length in pixel
 *   *scanlines number of scanlines
 */
static int vbe_scanline_get(unsigned* byte_length, unsigned* pixel_length, unsigned* scanlines) {
	__dpmi_regs r;

	assert( vbe_mode_active() );

	r.x.ax = 0x4F06;
	r.x.bx = 0x01;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F06/01 (GetScanlineLength)",(unsigned)r.x.ax);
		return r.x.ax;
	}

	*byte_length = r.x.bx;
	*pixel_length = r.x.cx;
	*scanlines = r.x.dx;

	return 0;
}

/* Set the length of the scanline in byte
 * return:
 *   0 on successfull
 * note:
 *   Update correctly      vbe_state
 *   Length is incremented to an acceptable value
 */
int vbe_scanline_set(unsigned byte_length) {
	__dpmi_regs r;

	assert( vbe_mode_active() );

	/* adjust */
	byte_length = vbe_adjust_scanline_size(byte_length);
	if (!byte_length) {
		video_error_description_set("Invalid byte length for vbe_scanline_set");
		return -1;
	}

	r.x.ax = 0x4F06;
	r.x.bx = 0x02; /* TODO VESA 1.2 does't support mode 0x02 */
	r.x.cx = byte_length;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F06/02 (SetScanlineLength)",(unsigned)r.x.ax);
		return r.x.ax;
	}

	/* update vbe_state */
	vbe_state.bytes_per_scanline = r.x.bx;
	vbe_state.virtual_x = r.x.cx;
	vbe_state.virtual_y = r.x.dx;

	return 0;
}

/* Set the length of the scanline in pixel
 * return:
 *   0 on successfull
 * note:
 *   Length is incremented to an acceptable value
 */
int vbe_scanline_pixel_set(unsigned pixel_length) {
	return vbe_scanline_set(pixel_length * vbe_state.bytes_per_pixel);
}

int vbe_scanline_pixel_request(unsigned pixel_length) {
	if (pixel_length > vbe_state.virtual_x) {
		if (vbe_scanline_pixel_set(pixel_length) != 0) {
			return -1;
		}
	}

	/* retest */
	if (pixel_length > vbe_state.virtual_x) {
		video_error_description_set("Unexpected virtual x size");
		return -1;
	}

	return 0;
}

/***************************************************************************/
/* Linear frame buffer */

static unsigned FAST vbe_linear_write_selector(void) {
	return vbe_state.linear_selector;
}

static unsigned FAST vbe_linear_write_offset(unsigned offset, unsigned size) {
	(void)size; /* prevent warning */
	return offset;
}

static unsigned FAST vbe_linear_write_line(unsigned y) {
	return y * vbe_state.bytes_per_scanline;
}

/* Inizialize the vbe linear framebuffer
 * return:
 *      0 on successfull
 */
static int vbe_linear_init(void) {
	assert( vbe_mode_active() );
	assert( !vbe_linear_active() );

	if (vbe_state.info.VESAVersion < 0x200) {
		video_error_description_set("Linear frame buffer is avaliable only with VBE2 or greather");
		return -1;
	}

	if (physical_init(&vbe_state.linear_selector, vbe_state.mode_info.PhysBasePtr, vbe_state.info.TotalMemory * 0x10000)!=0)
		return -1;

	/* enable write function */
	vbe_write_selector = vbe_linear_write_selector;
	vbe_write_line = vbe_linear_write_line;
	vbe_write_offset = vbe_linear_write_offset;

	/* enable */
	vbe_state.linear_active = 1;

	return 0;
}

/* Deinizialize the vbe linear framebuffer */
static void vbe_linear_done(void) {
	assert( vbe_linear_active() );

	physical_done(vbe_state.linear_selector,vbe_state.mode_info.PhysBasePtr, vbe_state.info.TotalMemory * 0x10000);

	/* disable write function */
	vbe_write_selector = 0;
	vbe_write_line = 0;
	vbe_write_offset = 0;

	/* disable */
	vbe_state.linear_active = 0;
}

/***************************************************************************/
/* Protect mode */

static __inline__ void vbe_pm_call(uint32 addr, uint32 ebx, uint32 ecx, uint32 edx) {
	/* From VBE 3.0 Specifications */
	/* Note: Currently undefined registers may be destroyed with the */
	/* exception of ESI, EBP, DS and SS. */
	__asm__ __volatile__ (
		"pushw %%es ;"
		"andl %%eax,%%eax ;"
		"jz 0f ;"
		"movw %%ax, %%es ;"
		"0: ;"
		"call *%%edi ;"
		"popw %%es ;"
		: /* no output */
		:
		"a" (vbe_state.pm_mmio_selector),
		"b" (ebx),
		"c" (ecx),
		"d" (edx),
		"D" (addr)
		:
#ifdef VBE_BOGUS
		"cc", "esi", "ebp"
#else
		"cc"
#endif
	);
}

static __inline__ void vbe_pm_call_ptr(uint32 addr, uint32 ebx, uint32 ecx, uint32 edx, uint32 es, uint32 edi) {
	__asm__ __volatile__ (
		"pushw %%es ;"
		"pushw %%ds ;"
		"movw %%ax, %%es ;"
		"movl %4,%%eax ;"
		"andl %%eax,%%eax ;"
		"jz 0f ;"
		"movw %%ax, %%ds ;"
		"0: ;"
		"call *%0 ;"
		"popw %%ds ;"
		"popw %%es ;"
		: /* no output */
		:
		"r" (addr),
		"b" (ebx),
		"c" (ecx),
		"d" (edx),
		"g" (vbe_state.pm_mmio_selector),
		"a" (es),
		"D" (edi)
		:
#ifdef VBE_BOGUS
		"cc" /*, "esi", "ebp" */
#else
		"cc"
#endif
	);
}

/* Inizialize the protect mode interface
 * return:
 *   ==0 ok
 * note:
 *   Call after the mode set
 */
static int vbe_pm_init(void) {
	__dpmi_regs r;
	unsigned len;

	assert( !vbe_pm_active() );
	assert( vbe_mode_active() );

	vbe_state.pm_switcher = 0;
	vbe_state.pm_scroller = 0;
	vbe_state.pm_palette = 0;
	vbe_state.pm_mmio_selector = 0;
	vbe_state.pm_mmio_address = 0;
	vbe_state.pm_mmio_size = 0;

	if (vbe_state.info.VESAVersion < 0x200) {
		video_error_description_set("Protect mode interface avaliable only with VBE2 or greather");
		return -1;
	}

	r.x.ax = 0x4F0A;
	r.x.bx = 0;
	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F0A/00 (GetProtectModeInterface)",(unsigned)r.x.ax);
		return r.x.ax;
	}

	len = r.x.cx;
	vbe_state.pm_info = malloc(len);
	assert( vbe_state.pm_info );

	dosmemget(r.x.es*16+r.x.di, len, vbe_state.pm_info);

#ifdef VBE_BOGUS
	/* Now do a sanity check on the information we recieve to ensure */
	/* that is is correct. Some BIOS return totally bogus information */
	/* in here (some Matrox, Trident Blade3D) */
	if (vbe_state.pm_info->setWindow >= len ||
		vbe_state.pm_info->setDisplayStart >= len ||
		vbe_state.pm_info->setPalette >= len ||
		vbe_state.pm_info->IOPrivInfo >= len) {
		if (vbe_state.pm_info->setWindow >= len)
			video_error_description_set("Invalid setWindow addr");
		if (vbe_state.pm_info->setDisplayStart >= len)
			video_error_description_set("Invalid setDisplayStart addr");
		if (vbe_state.pm_info->setPalette >= len)
			video_error_description_set("Invalid setPalette addr");
		if (vbe_state.pm_info->IOPrivInfo >= len)
			video_error_description_set("Invalid IOPrivInfo addr");
		video_error_description_set("Inconsistent information from the VESA call 4F0A/00 (GetProtectModeInterface)");

		free(vbe_state.pm_info);
		vbe_state.pm_info = 0;
		return -1;
	}
#endif

	/* memory mapped IO */
	if (vbe_state.pm_info->IOPrivInfo) {
		uint16* p;
		p = (uint16*)((uint8*)vbe_state.pm_info + vbe_state.pm_info->IOPrivInfo);
		/* skip the port table */
		while (*p != 0xFFFF)
			p++;
		p++;
		if (*p != 0xFFFF) {
			vbe_state.pm_mmio_address = *((uint32*)p);
			vbe_state.pm_mmio_size = *(p+2);
			if (physical_init(&vbe_state.pm_mmio_selector, vbe_state.pm_mmio_address, vbe_state.pm_mmio_size ) != 0) {
				vbe_state.pm_mmio_selector = 0;
				vbe_state.pm_mmio_address = 0;
				vbe_state.pm_mmio_size = 0;
				return -1;
			}
		}
	}

	/* assume cs==ds */
	vbe_state.pm_switcher = (unsigned)vbe_state.pm_info + vbe_state.pm_info->setWindow;
	vbe_state.pm_scroller = (unsigned)vbe_state.pm_info + vbe_state.pm_info->setDisplayStart;
	vbe_state.pm_palette = (unsigned)vbe_state.pm_info + vbe_state.pm_info->setPalette;

	/* enable */
	vbe_state.pm_active = 1;

	return 0;
}

static void vbe_pm_done(void) {
	assert( vbe_pm_active() );

	free( vbe_state.pm_info );

	if (vbe_state.pm_mmio_selector) {
		physical_done(vbe_state.pm_mmio_selector, vbe_state.pm_mmio_address, vbe_state.pm_mmio_size);
	}

	/* disable */
	vbe_state.pm_active = 0;
}

/***************************************************************************/
/* Banked frame buffer */

/* Select one bank
 * in:
 *   offset in bytes
 * return:
 *      0 on successfull
 */
static int vbe_bank(unsigned offset) {
	unsigned wingpos = (offset >> 10) / vbe_state.mode_info.WinGranularity;

	if (vbe_state.pm_active && vbe_state.pm_switcher) {
		vbe_pm_call(vbe_state.pm_switcher, vbe_state.bank_winnumber, 0, wingpos);
	} else {
		__dpmi_regs r;

		r.x.ax = 0x4F05;
		r.x.bx = vbe_state.bank_winnumber;
		r.x.dx = wingpos;

		__dpmi_int(0x10, &r);
		if (r.x.ax!=0x004F) {
			video_error_description_set("Error %x in the VESA call 4F05 (SetBank)",(unsigned)r.x.ax);
			return r.x.ax;
		}
	}

	return 0;
}

static unsigned FAST vbe_bank_write_selector(void) {
	return vbe_state.bank_selector;
}

static unsigned FAST vbe_bank_write_offset(unsigned offset, unsigned size) {
	if (offset < vbe_state.bank_begin || offset + size >= vbe_state.bank_end) {
		unsigned mask = vbe_state.mode_info.WinGranularity;
		mask = mask * 1024 - 1;
		vbe_state.bank_begin = offset & ~mask;
		vbe_state.bank_end = vbe_state.bank_begin + vbe_state.mode_info.WinSize * 1024;
		vbe_bank( vbe_state.bank_begin );
	}
	return vbe_state.bank_offset + offset - vbe_state.bank_begin;
}

static unsigned FAST vbe_bank_write_line(unsigned y) {
	return vbe_bank_write_offset(y * vbe_state.bytes_per_scanline, vbe_state.bytes_per_scanline);
}

/* Inizialize
 * return:
 *      0 on successfull
 */
static int vbe_bank_init(void) {
	assert( vbe_mode_active() );
	assert( !vbe_bank_active() );

	if (vbe_state.mode_info.WinAAttributes & 0x4) {
		vbe_state.bank_selector = _dos_ds;
		vbe_state.bank_offset = vbe_state.mode_info.WinASegment * 16;
		vbe_state.bank_winnumber = 0;
	} else if (vbe_state.mode_info.WinBAttributes & 0x4) {
		vbe_state.bank_selector = _dos_ds;
		vbe_state.bank_offset = vbe_state.mode_info.WinBSegment * 16;
		vbe_state.bank_winnumber = 1;
	} else {
		video_error_description_set("Invalid bank descpription");
		return -1;
	}

	/* check that all scanlines are addressable */
	if (vbe_state.mode_info.WinSize == vbe_state.mode_info.WinGranularity &&
		vbe_state.mode_info.WinSize * 1024 % vbe_state.bytes_per_scanline != 0) {
		/* reset the scanline size, adjusted to an acceptable value */
		unsigned bytes_per_scanline_large = vbe_bank_adjust_scanline_size(vbe_state.bytes_per_scanline);
		if (vbe_scanline_set(bytes_per_scanline_large)!=0)
			return -1;
	}

	/* select bank at offset 0 */
	vbe_bank(0);
	vbe_state.bank_begin = 0;
	vbe_state.bank_end = vbe_state.bank_begin + vbe_state.mode_info.WinSize * 1024;

	/* enable write function */
	vbe_write_selector = vbe_bank_write_selector;
	vbe_write_line = vbe_bank_write_line;
	vbe_write_offset = vbe_bank_write_offset;

	/* enable */
	vbe_state.bank_active = 1;

	return 0;
}

/* Deinizialize */
static void vbe_bank_done(void) {
	assert( vbe_bank_active() );

	/* disable */
	vbe_state.bank_active = 0;
}

/***************************************************************************/
/* text */

/* Return the selector for accessing in writing the video memory */
static unsigned FAST vbe_text_write_selector(void) {
	return _dos_ds;
}

/* Return the offset for accessing in writing the video memory */
static unsigned FAST vbe_text_write_line(unsigned y) {
	assert( vbe_mode_active() );
	return 0xA0000 + y * vbe_state.bytes_per_scanline;
}

/***************************************************************************/
/* Mode */

void vbe_mode_done(void) {
	assert( vbe_mode_active() );

	if (vbe_linear_active())
		vbe_linear_done();
	if (vbe_bank_active())
		vbe_bank_done();
	if (vbe_pm_active() )
		vbe_pm_done();

	/* disable */
	vbe_state.mode_active = 0;
}

#if 0 /* not used */
static int vbe_dac_set(unsigned width) {
	__dpmi_regs r;

	assert( vbe_active() && vbe_mode_active() );

	r.x.ax = 0x4F08;
	r.h.bl = 0x00;
	r.h.bh = width;
	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F08 (SetPalette)",(unsigned)r.x.ax);
		return r.x.ax;
	}

	vbe_state.palette_width = r.h.bh;
	video_log("vbe: palette set to %d bit\n",(unsigned)vbe_state.palette_width);

	return 0;
}
#endif

/* Compute shift and mask value */
static void vbe_rgb_data(int* shift, unsigned* mask, unsigned posbit, unsigned masksize) {
	*mask = ((1 << masksize) - 1) << posbit;
	*shift = posbit + masksize - 8;
}

/* Set one graphics mode
 * in:
 *   mode mode number with vbeClear or vbeLinear flags
 *   crtc CRTC information, ==0 for default
 * return:
 *   0 on successfull
 * note:
 *   update correctly      vbe_state
 *   inizialize linear frambuffer if requested
 *   if fail, video mode can be bronken
 */
int vbe_mode_set(unsigned mode, const vbe_CRTCInfoBlock* crtc) {
	vbe_ModeInfoBlock info;
	__dpmi_regs r;

	video_log("vbe: mode_set %x\n",mode);

	assert( vbe_active() );

	/* check support for CRTC */
	if (crtc && vbe_state.info.VESAVersion < 0x300) {
		video_error_description_set("CRTC programming is available only with VBE3 or greather");
		return -1;
	}

	/* check support for linear frame buffer */
	if ((mode & vbeLinearBuffer)!=0 && vbe_state.info.VESAVersion < 0x200) {
		video_error_description_set("Linear Frame Buffer is available only with VBE2 or greather");
		return -1;
	}

	/* get mode info */
	if (vbe_mode_info_get(&info,mode)!=0) {
		return -1;
	}

	if (vbe_mode_active())
		vbe_mode_done();

	r.x.ax = 0x4F02;
	if (crtc) {
		dosmemput(crtc, sizeof(vbe_CRTCInfoBlock), __tb);
		r.x.bx = mode | vbeRefreshCtrl;
		r.x.es = __tb / 16;
		r.x.di = 0;
	} else {
		r.x.bx = mode;
	}

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F02/%x (ModeSet)",(unsigned)r.x.ax,(unsigned)mode);
		return r.x.ax;
	}

	/* enable */
	vbe_state.mode_active = 1;

	/* update vbe_state.mode_info */
	vbe_state.mode_info = info;

	/* update vbe_state */
	vbe_state.mode = mode;
	vbe_state.scroll = 0;

	if (vbe_state.mode_info.ModeAttributes & vbeMdGraphMode) {
		/* graphics mode */
		vbe_state.size_x = vbe_state.mode_info.XResolution;
		vbe_state.size_y = vbe_state.mode_info.YResolution;

		vbe_state.bytes_per_pixel = (vbe_state.mode_info.BitsPerPixel + 7) / 8;
		if (vbe_scanline_get(&vbe_state.bytes_per_scanline, &vbe_state.virtual_x, &vbe_state.virtual_y)!=0) {
#ifdef VBE_BOGUS
			/* Retry setting */
			if (vbe_scanline_pixel_set(info.XResolution)!=0) {
				vbe_mode_done();
				return -1;
			}
#else
			vbe_mode_done();
			return -1;
#endif
		}

		if (vbe_state.mode_info.MemoryModel==vbeMemRGB || vbe_state.mode_info.MemoryModel==vbeMemYUV) {
			/* update color rgb */
			if (mode & vbeLinearBuffer) {
				vbe_rgb_data(&vbe_state.rgb_red_shift,&vbe_state.rgb_red_mask,vbe_state.mode_info.RedFieldPosition,vbe_state.mode_info.RedMaskSize);
				vbe_rgb_data(&vbe_state.rgb_green_shift,&vbe_state.rgb_green_mask,vbe_state.mode_info.GreenFieldPosition,vbe_state.mode_info.GreenMaskSize);
				vbe_rgb_data(&vbe_state.rgb_blue_shift,&vbe_state.rgb_blue_mask,vbe_state.mode_info.BlueFieldPosition,vbe_state.mode_info.BlueMaskSize);
			} else {
				vbe_rgb_data(&vbe_state.rgb_red_shift,&vbe_state.rgb_red_mask,vbe_state.mode_info.LinRedFieldPosition,vbe_state.mode_info.LinRedMaskSize);
				vbe_rgb_data(&vbe_state.rgb_green_shift,&vbe_state.rgb_green_mask,vbe_state.mode_info.LinGreenFieldPosition,vbe_state.mode_info.LinGreenMaskSize);
				vbe_rgb_data(&vbe_state.rgb_blue_shift,&vbe_state.rgb_blue_mask,vbe_state.mode_info.LinBlueFieldPosition,vbe_state.mode_info.LinBlueMaskSize);
			}
		} else {
			/* update color palette */
			vbe_state.palette_width = 6; /* VBE default */
#if 0 /* disabled for no testing possibility */
			if ((vbe_state.info.Capabilities & vbe8BitDAC) != 0) {
				video_log("vbe: try setting 8 bit palette\n");
				if (vbe_dac_set(8)!=0) {
					video_log("vbe: set 8 bit palette FAILED\n");
					/* ignore error, if failed the default is 6 bit */
				}
			}
#endif
		}

		/* inizialize pm interface */
		if (vbe_state.info.VESAVersion >= 0x200) {
			if (vbe_pm_init() != 0) {
				/* ignore error, don't use pm interface */
			}
		}

		if (mode & vbeLinearBuffer) {
			/* inizialize linear framebuffer */
			if (vbe_linear_init() != 0) {
				vbe_mode_done();
				return -1;
			}
		} else {
			/* inizialize banked framebuffer */
			if (vbe_bank_init() != 0) {
				vbe_mode_done();
				return -1;
			}
		}
	} else {
		/* text mode */
		vbe_state.size_x = vbe_state.mode_info.XResolution * vbe_state.mode_info.XCharSize;
		vbe_state.size_y = vbe_state.mode_info.YResolution * vbe_state.mode_info.YCharSize;
		vbe_state.virtual_x = vbe_state.size_x;
		vbe_state.virtual_y = vbe_state.virtual_y;

		vbe_state.bytes_per_pixel = 0;
		vbe_state.bytes_per_scanline = vbe_state.mode_info.XResolution * 2;
		vbe_state.palette_width = 0;
		vbe_write_selector = vbe_text_write_selector;
		vbe_write_line = vbe_text_write_line;
		vbe_write_offset = 0;
	}

	return 0;
}

/* Get the current graphics mode
 */
int vbe_mode_get(unsigned* mode) {
	__dpmi_regs r;

	assert( vbe_active() );

	r.x.ax = 0x4F03;
	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F03 (ModeGet)",(unsigned)r.x.ax);
		return r.x.ax;
	}

	*mode = r.x.bx & 0x7FFF; /* clear unused bits */

	return 0;
}

/***************************************************************************/
/* VBE */

static void dosstrzcpy(char* dst, unsigned segoff, unsigned len) {
	unsigned ptr = ((segoff & 0xFFFF0000) >> 12) + (segoff & 0xFFFF);

	if (!segoff) {
		*dst = 0;
		return;
	}

	while (len > 1 && _farpeekb(_dos_ds, ptr) != 0x00) {
		*dst = _farpeekb(_dos_ds, ptr);
		++dst;
		--len;
		++ptr;
	}
	*dst = 0;
}

static void dosmodezcpy(unsigned* dst, unsigned segoff, unsigned len) {
	unsigned ptr = ((segoff & 0xFFFF0000) >> 12) + (segoff & 0xFFFF);

	if (!segoff) {
		*dst = 0xFFFF;
		return;
	}

	while (len > 1 && _farpeekw(_dos_ds, ptr) != 0xFFFF) {
		*dst = _farpeekw(_dos_ds, ptr);
		++dst;
		--len;
		ptr += 2;
	}
	*dst = 0xFFFF;
}

/* Inizialize the vbe system
 * return:
 *      0 on successfull
 */
int vbe_init(void) {
	__dpmi_regs r;
	char oem_string[256];
	char oem_vendor[256];
	char oem_product[256];
	char oem_product_rev[256];
	unsigned i;

	assert( !vbe_active() );

	/* inizialize with all zero */
	memset(&vbe_state.info, 0, sizeof(vbe_VbeInfoBlock) );

	/* request VBE 2.0+ */
	memcpy(&vbe_state.info.VESASignature,"VBE2",4);

	dosmemput(&vbe_state.info, sizeof(vbe_VbeInfoBlock), __tb);

	r.x.ax = 0x4F00;
	r.x.es = __tb / 16;
	r.x.di = 0;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		unsigned char sign[4];
		dosmemget(__tb, 4, &sign);
		video_log("vbe: not found, status %04x, sign %02x%02x%02x%02x \n",(unsigned)r.x.ax, (unsigned)sign[0],(unsigned)sign[1],(unsigned)sign[2],(unsigned)sign[3]);
		video_error_description_set("Error %x in the VESA call 4F00 (GetVesaInfo)",(unsigned)r.x.ax);
		return -1;
	}

	dosmemget(__tb, sizeof(vbe_VbeInfoBlock), &vbe_state.info);

	/* reject too old bios */
	if (vbe_state.info.VESAVersion < 0x102) {
		video_log("vbe: version lower than 1.2\n");
		video_error_description_set("The VESA BIOS is of a version prior 1.2");
		return -1;
	}

#ifdef VBE_BOGUS
	/* check for bogus BIOSes */
	if (vbe_state.info.VESAVersion >= 0x200 && vbe_state.info.OemVendorNamePtr == 0)
		vbe_state.info.VESAVersion = 0x102;
#endif

	/* bufferize all the dynamic info */
	dosmodezcpy(vbe_mode_map,vbe_state.info.VideoModePtr,VBE_MODE_MAX);
	dosstrzcpy(oem_string,vbe_state.info.OemStringPtr,sizeof(oem_string));
	if (vbe_state.info.VESAVersion >= 0x200) {
		dosstrzcpy(oem_vendor,vbe_state.info.OemVendorNamePtr,sizeof(oem_vendor));
		dosstrzcpy(oem_product,vbe_state.info.OemProductNamePtr,sizeof(oem_product));
		dosstrzcpy(oem_product_rev,vbe_state.info.OemProductRevPtr,sizeof(oem_product_rev));
	} else {
		strcpy(oem_vendor,"");
		strcpy(oem_product,"");
		strcpy(oem_product_rev,"");
	}

	video_log("vbe: version %d.%d\n",(unsigned)vbe_state.info.VESAVersion >> 8, (unsigned)vbe_state.info.VESAVersion & 0xFF);
	video_log("vbe: memory %d\n", (unsigned)vbe_state.info.TotalMemory * 0x10000);
	if (vbe_state.info.Capabilities & vbeNonVGA)
		video_log("vbe: controller is NOT VGA compatible\n");
	else
		video_log("vbe: controller is VGA compatible\n");

	video_log("vbe: oem %s\n",oem_string);

	if (vbe_state.info.VESAVersion >= 0x200) {
		video_log("vbe: oem software rev %d.%d\n",vbe_state.info.OemSoftwareRev >> 8,vbe_state.info.OemSoftwareRev & 0xFF);
		video_log("vbe: oem vendor name ptr %s\n",oem_vendor);
		video_log("vbe: oem product name ptr %s\n",oem_product);
		video_log("vbe: oem product rev ptr %s\n",oem_product_rev);
	}

	/* enable */
	vbe_state.active = 1;

	/* print modes */
	for(i=0;vbe_mode_map[i]!=0xFFFF;++i) {
		vbe_ModeInfoBlock info;
		if (vbe_mode_info_get(&info, vbe_mode_map[i])!=0) {
			video_log("vbe: error getting info for mode %x\n",(unsigned)vbe_mode_map[i]);
		} else {
			if (!(info.ModeAttributes & vbeMdGraphMode)) {
				video_log("vbe: mode %4xh %4dx%4d text\n",vbe_mode_map[i],info.XResolution,info.YResolution);
			} else if (info.MemoryModel != vbeMemPK && info.MemoryModel != vbeMemRGB) {
				video_log("vbe: mode %4xh %4dx%4d %2d not packed/rgb\n",vbe_mode_map[i],info.XResolution,info.YResolution,info.BitsPerPixel);
			} else if (info.NumberOfPlanes > 1) {
				video_log("vbe: mode %4xh %4dx%4d %2d planar\n",vbe_mode_map[i],info.XResolution,info.YResolution,info.BitsPerPixel);
			} else {
				const char* flag_doublescan = info.ModeAttributes & vbeMdDoubleScan ? " doublescan": "";
				const char* flag_interlace = info.ModeAttributes & vbeMdInterlace ? " interlace": "";
				const char* flag_triple = info.ModeAttributes & vbeMdTripleBuffer ? " triple": "";
				const char* flag_linear = info.ModeAttributes & vbeMdLinear ? " linear": "";
				const char* flag_banked = !(info.ModeAttributes & vbeMdNonBanked) ? " banked": "";
				const char* flag_vga = !(info.ModeAttributes & vbeMdNonVGA) ? " vga": "";
				video_log("vbe: mode %4xh %4dx%4d %2d%s%s%s%s%s%s \n",vbe_mode_map[i],info.XResolution,info.YResolution,info.BitsPerPixel,flag_doublescan,flag_interlace,flag_triple,flag_linear,flag_banked,flag_vga);
			}
		}
	}

	return 0;
}

/* Deinizialize the vbe system */
void vbe_done(void) {
	assert( vbe_active() );

	if (vbe_mode_active()) {
		vbe_mode_done();
		vga_mode_set(0x3,0,0);
	}

	assert( !vbe_mode_active() );
	assert( !vbe_linear_active() );
	assert( !vbe_bank_active() );
	assert( !vbe_pm_active() );

	/* disable */
	vbe_state.active = 0;
}

unsigned vbe_adjust_bytes_per_page(unsigned bytes_per_page) {
	return bytes_per_page = (bytes_per_page + 0xFFFF) & ~0xFFFF;
}

/* Get information for a vbe mode
 * return:
 *   0 on successfull
 * note:
 *   VBE 3.0 information are always filled, also if VBE is less than 3.0
 */
int vbe_mode_info_get(vbe_ModeInfoBlock* info, unsigned mode) {
	__dpmi_regs r;

	assert( vbe_active() );

	r.x.ax = 0x4F01;
	r.x.cx = mode & vbeModeMask;
	r.x.es = __tb / 16;
	r.x.di = 0;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F01/%x (GetModeInfo)",(unsigned)r.x.ax,(unsigned)mode);
		return -1;
	}

	dosmemget( __tb, sizeof(vbe_ModeInfoBlock), info);

#ifdef VBE_BOGUS
	/* support old VESA implementation */
	if (vbe_state.info.VESAVersion < 0x200) {
		
		/* convert strange mode */
		if (info->MemoryModel == vbeMemPK && info->BitsPerPixel > 8) {
			switch (info->BitsPerPixel) {
				case 15:
					info->MemoryModel = vbeMemRGB;
					info->RedMaskSize = 5;
					info->RedFieldPosition = 10;
					info->GreenMaskSize = 5;
					info->GreenFieldPosition = 5;
					info->BlueMaskSize = 5;
					info->BlueFieldPosition = 0;
					info->RsvdMaskSize = 1;
					info->RsvdFieldPosition = 15;
					break;
				case 16:
					info->MemoryModel = vbeMemRGB;
					info->RedMaskSize = 5;
					info->RedFieldPosition = 11;
					info->GreenMaskSize = 5;
					info->GreenFieldPosition = 5;
					info->BlueMaskSize = 5;
					info->BlueFieldPosition = 0;
					info->RsvdMaskSize = 0;
					info->RsvdFieldPosition = 0;
				break;
				case 24:
					info->MemoryModel = vbeMemRGB;
					info->RedMaskSize = 8;
					info->RedFieldPosition = 16;
					info->GreenMaskSize = 8;
					info->GreenFieldPosition = 8;
					info->BlueMaskSize = 8;
					info->BlueFieldPosition = 0;
					info->RsvdMaskSize = 0;
					info->RsvdFieldPosition = 0;
					break;
			}
		}

		/* detect 15 bits info */
		if (info->BitsPerPixel==16 && info->RsvdMaskSize == 1)
			info->BitsPerPixel = 15;
	}
#endif

	/* extend information to VBE 3.0 format */
	if (vbe_state.info.VESAVersion < 0x300) {
		info->LinBytesPerScanLine = info->BytesPerScanLine;
		info->BnkNumberOfPages = info->NumberOfImagePages;
		info->LinNumberOfPages = info->NumberOfImagePages;
		info->LinRedMaskSize = info->RedMaskSize;
		info->LinRedFieldPosition = info->RedFieldPosition;
		info->LinGreenMaskSize = info->GreenMaskSize;
		info->LinGreenFieldPosition = info->GreenFieldPosition;
		info->LinBlueMaskSize = info->BlueMaskSize;
		info->LinBlueFieldPosition = info->BlueFieldPosition;
		info->LinRsvdMaskSize = info->RsvdMaskSize;
		info->LinRsvdFieldPosition = info->RsvdFieldPosition;
		info->MaxPixelClock = 0;
	}
	return 0;
}

int vga_as_vbe_mode_info_get(vbe_ModeInfoBlock* info, unsigned mode) {
	memset(info,0,sizeof(vbe_ModeInfoBlock));
	if (mode == 0x13) {
		info->ModeAttributes = vbeMdAvailable | vbeMdTTYOutput | vbeMdColorMode | vbeMdGraphMode | vbeMdDoubleScan;
		info->BytesPerScanLine = 320;
		info->XResolution = 320;
		info->YResolution = 200;
		info->NumberOfPlanes = 1;
		info->BitsPerPixel = 8;
		info->MemoryModel = vbeMemPK;
	} else if (mode == 0x3) {
		info->ModeAttributes = vbeMdAvailable | vbeMdTTYOutput | vbeMdColorMode | vbeMdDoubleScan;
		info->BytesPerScanLine = 160;
		info->XResolution = 80;
		info->YResolution = 25;
		info->XCharSize = 9;
		info->YCharSize = 16;
		info->NumberOfPlanes = 1;
		info->BitsPerPixel = 4;
		info->MemoryModel = vbeMemTXT;
	} else {
      		video_error_description_set("VBE emulation only for mode VGA 13h,03h");
		return -1;
	}

	return 0;
}


int vbe_scroll(unsigned offset, int waitvsync) {
	assert( vbe_active() && vbe_mode_active() );

#ifndef VBE_BOGUS
	/* On my clean hardware always wait a vsync, but with software
	driver work ok */
	if (vbe_state.pm_active && vbe_state.pm_scroller) {
		unsigned bx,cx,dx;
		bx = waitvsync ? 0x80 : 0x0;
		/* assume no planar mode (don't divide offset by 4) */
		cx = (offset >> 2) & 0xFFFF;
		dx = offset >> 18;
		/* use lower 2 bit if supported */
		if (vbe_state.info.VESAVersion >= 0x300) {
			dx |= (offset & 0x3) << 14;
		}
		vbe_pm_call(vbe_state.pm_scroller, bx, cx, dx);
	} else
#endif
	{
		__dpmi_regs r;

		r.x.ax = 0x4F07;
		if (vbe_state.info.VESAVersion >= 0x300) {
			r.x.bx = waitvsync ? 0x82 : 0x2;
			r.d.ecx = offset;
		} else {
			r.x.bx = waitvsync ? 0x80 : 0x0;
			r.x.cx = (offset % vbe_state.bytes_per_scanline) / vbe_state.bytes_per_pixel;
			r.x.dx = offset / vbe_state.bytes_per_scanline;
		}
		__dpmi_int(0x10, &r);
		if (r.x.ax!=0x004F) {
			video_error_description_set("Error %x in the VESA call 4F07 (ScrollSet)",(unsigned)r.x.ax);
			return -1;
		}
	}

	vbe_state.scroll = offset;

	return 0;
}

int vbe_request_scroll(unsigned offset) {
	__dpmi_regs r;

	assert( vbe_active() && vbe_mode_active() );

	r.x.ax = 0x4F07;
	r.x.bx = 0x02;
	r.d.ecx = offset;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F07/02 (ScrollRequest)",(unsigned)r.x.ax);
		return -1;
	}

	vbe_state.scroll = offset;

	return 0;
}

/*
 * note:
 *   This function is not supported by protected mode interface
 */
int vbe_poll_scroll(unsigned* done) {
	__dpmi_regs r;

	assert( vbe_active() && vbe_mode_active() );

	r.x.ax = 0x4F07;
	r.x.bx = 0x04;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F07/04 (ScrollPoll)",(unsigned)r.x.ax);
		return -1;
	}
	
	*done = r.h.bl;

	return 0;
}

/* Wait a vsync */
void vbe_wait_vsync(void) {
	assert( vbe_active() && vbe_mode_active() );

	/* if possible use vga version */
	if ((vbe_state.info.Capabilities & vbeNonVGA)==0) {
		vga_wait_vsync();
	} else {
		vbe_scroll(vbe_state.scroll,1);
	}
}

/* Value used for incrementing/decrementin pixelclock */
#define VBE_PIXELCLOCK_DELTA (10*1000)
#define VBE_PIXELCLOCK_MAX (250*1000*1000)
#define VBE_PIXELCLOCK_MIN (1*1000*1000)

/* Return the next avaliable pixel clock */
int vbe_pixelclock_getnext(unsigned* pixelclock, unsigned mode) {
	unsigned newpixelclock = *pixelclock + VBE_PIXELCLOCK_DELTA;
	while (newpixelclock < VBE_PIXELCLOCK_MAX) {
		unsigned testpixelclock = newpixelclock;
		if (vbe_pixelclock_get(&testpixelclock,mode)!=0) {
			return -1;
		}
		if (testpixelclock > *pixelclock) {
			*pixelclock = testpixelclock;
			return 0;
		}
		newpixelclock += VBE_PIXELCLOCK_DELTA;
	}
	return -1;
}

/* Return the pred avaliable pixel clock */
int vbe_pixelclock_getpred(unsigned* pixelclock, unsigned mode) {
	unsigned newpixelclock = *pixelclock - VBE_PIXELCLOCK_DELTA;
	while (newpixelclock > VBE_PIXELCLOCK_MIN) {
		unsigned testpixelclock = newpixelclock;
		if (vbe_pixelclock_get(&testpixelclock,mode)!=0) {
			return -1;
		}
		if (testpixelclock < *pixelclock) {
			*pixelclock = testpixelclock;
			return 0;
		}
		newpixelclock -= VBE_PIXELCLOCK_DELTA;
	}
	return -1;
}

/* Get a valid pixel clock
 * in:
 *   *pixelclock to test in Hz
 *   mode mode number (with vbeLinearBuffer if used for)
 * out:
 *   *pixelclock nearest pixelclock in Hz
 */
int vbe_pixelclock_get(unsigned* pixelclock, unsigned mode) {
	__dpmi_regs r;

	assert( vbe_active() );

	r.x.ax = 0x4F0B;
	r.x.bx = 0x00;
	r.d.ecx = *pixelclock;
	r.x.dx = mode & vbeModeMask;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F0B/00 (PixelClockGet)",(unsigned)r.x.ax);
		return -1;
	}

	*pixelclock = r.d.ecx;

	return 0;
}

/* Set palette
 * in:
 *   palette palette to set in DAC format
 * note:
 *   Prefferred are not real-protected switch required method
 */
int vbe_palette_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	unsigned mode = waitvsync ? 0x80 : 0x0;

	assert( vbe_active() && vbe_mode_active() );

	if (vbe_state.pm_active && vbe_state.pm_palette) {
		/* video_log("vbe: palette set (%d bit) with VBE pm call\n",(unsigned)vbe_state.palette_width); */
		vbe_pm_call_ptr(vbe_state.pm_palette, mode, count, start, _my_ds(), (unsigned)palette);
	} else if ((vbe_state.info.Capabilities & vbeNonVGA)==0 && vbe_state.palette_width==6) {
		/* video_log("vbe: palette set (6 bit) with VGA registers\n"); */
		vga_palette6_set(palette,start,count,waitvsync);
#if 0 /* if the palette is set to 8 bit the BIOS call must be used */
	} else if ((vbe_state.info.Capabilities & vbeNonVGA)==0 && vbe_state.palette_width==8) {
		/* video_log("vbe: palette set (8 bit) with VGA registers\n"); */
		vga_palette8_set(palette,start,count,waitvsync);
#endif
	} else {
		__dpmi_regs r;

		/* video_log("vbe: palette set (%d bit) with VBE bios\n",(unsigned)vbe_state.palette_width); */

		dosmemput(palette, count * sizeof(video_color), __tb);

		r.x.ax = 0x4F07;
		r.x.bx = mode;
		r.x.cx = count;
		r.x.dx = start;
		r.x.es = __tb / 16;
		r.x.di = 0;

		__dpmi_int(0x10, &r);
		if (r.x.ax!=0x004F) {
			video_error_description_set("Error %x in the VESA call 4F07 (PaletteSet)",(unsigned)r.x.ax);
			return -1;
		}
	}

	return 0;
}

void vbe_mode_iterator_begin(vbe_mode_iterator* vmi) {
	assert( vbe_active() );
	vmi->mode_ptr = 0;
}

unsigned vbe_mode_iterator_get(vbe_mode_iterator* vmi) {
	return vbe_mode_map[vmi->mode_ptr];
}

int vbe_mode_iterator_end(vbe_mode_iterator* vmi) {
	return vbe_mode_map[vmi->mode_ptr] == 0xFFFF;
}

void vbe_mode_iterator_next(vbe_mode_iterator* vmi) {
	vmi->mode_ptr += 1;
}

int vbe_save_state(unsigned* psize, void** pptr) {
	__dpmi_regs r;
	unsigned size;
	void* ptr;

	assert( vbe_active() );

	r.x.ax = 0x4F04;
	r.x.cx = 0xF;
	r.h.dl = 0x00;
	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F04/00 (SaveRestoreState)",(unsigned)r.x.ax);
		return -1;
	}

	size = r.x.bx * 64;
	if (size > 4096)
		video_log( "vbe: WARNING! save state size %d, probably to big for the DJGPP real mode buffer\n",size);

	r.x.ax = 0x4F04;
	r.x.cx = 0xF;
	r.h.dl = 0x01;
	r.x.es = __tb / 16;
	r.x.bx = 0;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F04/01 (SaveRestoreState)",(unsigned)r.x.ax);
		return -1;
	}

	ptr = (void*)malloc(size);
	dosmemget(__tb, size, ptr);

	*psize = size;
	*pptr = ptr;
	
	return 0;
}

int vbe_restore_state(unsigned size, void* ptr) {
	__dpmi_regs r;

	assert( vbe_active() );

	dosmemput(ptr, size, __tb);

	r.x.ax = 0x4F04;
	r.x.cx = 0xF;
	r.h.dl = 0x02;
	r.x.es = __tb / 16;
	r.x.bx = 0;

	__dpmi_int(0x10, &r);
	if (r.x.ax!=0x004F) {
		video_error_description_set("Error %x in the VESA call 4F04/02 (SaveRestoreState)",(unsigned)r.x.ax);
		return -1;
	}

	return 0;
}

