/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 */

#include "compil.h"
#include "card.h"

#ifdef __PCIREAL__
/* If defined remap all MMIO address in the first 1M of memory */
#define PCI_ADDRESS_REMAP_TO_1M
#endif

/**************************************************************************/
/* mapping */

#ifdef __PCIREAL__

int card_create_linear_mapping(unsigned long *linear, unsigned long physaddr, int size)
{
	if (physaddr + size > 1024L*1024L)
		return -1;
	*linear = physaddr;
	return 0;
}

void card_remove_linear_mapping(unsigned long *linear)
{
	*linear = 0;
}

int card_create_selector(unsigned *segment, unsigned long linear, int size)
{
	if (linear + size > 1024L*1024L)
		return -1;
	*segment = linear >> 4;
	return 0;
}

void card_remove_selector(unsigned *segment)
{
	*segment = 0;
}

#else

#include <dpmi.h>

/* card_create_linear_mapping:
 *  Maps a physical address range into linear memory.
 */
int card_create_linear_mapping(unsigned long *linear, unsigned long physaddr, int size)
{
	__dpmi_meminfo meminfo;

	if (physaddr >= 0x100000) {
		/* map into linear memory */
		meminfo.address = physaddr;
		meminfo.size = size;
		if (__dpmi_physical_address_mapping(&meminfo) != 0)
			return -1;

		*linear = meminfo.address;

		/* lock the linear memory range */
		__dpmi_lock_linear_region(&meminfo);
	} else {
		/* exploit 1 -> 1 physical to linear mapping in low megabyte */
		*linear = physaddr;
	}
	return 0;
}

/* card_remove_linear_mapping:
 *  Frees the DPMI resources being used to map a linear address range.
 */
void card_remove_linear_mapping(unsigned long *linear)
{
	__dpmi_meminfo meminfo;

	if (*linear) {
		if (*linear >= 0x100000) {
			meminfo.address = *linear;
			__dpmi_free_physical_address_mapping(&meminfo);
		}

		*linear = 0;
	}
}

/* card_create_selector:
 *  Allocates a selector to access a region of linear memory.
 */
int card_create_selector(unsigned *segment, unsigned long linear, int size)
{
	/* allocate an ldt descriptor */
	int des = __dpmi_allocate_ldt_descriptors(1);
	if (des < 0) {
		*segment = 0;
		return -1;
	}
	*segment = des;

	/* set the descriptor base and limit */
	__dpmi_set_segment_base_address(*segment, linear);

	--size;
	if (size < 0xFFFF)
		size = 0xFFFF;
	__dpmi_set_segment_limit(*segment, size);

	return 0;
}

/* card_remove_selector:
 *  Frees a DPMI segment selector.
 */
void card_remove_selector(unsigned *segment)
{
	if (*segment) {
		__dpmi_free_ldt_descriptor(*segment);
		*segment = 0;
	}
}

#endif

/***************************************************************************/
/* PCI */

int pci_detect(void) {
	__dpmi_regs r;
	r.d.eax = 0x0000b101;
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if (r.d.edx != 0x20494350)
		return -1;
	return 0;
}

int pci_find_device(unsigned vendor, unsigned device, unsigned STACK_PTR * bus_device_func) {
	__dpmi_regs r;
	r.d.eax = 0x0000b102;
	r.d.ecx = device;
	r.d.edx = vendor;
	r.d.esi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if (r.h.ah != 0)
		return -1;
	*bus_device_func = r.d.ebx;
	return 0;
}

int pci_read_dword(unsigned bus_device_func, unsigned reg, DWORD STACK_PTR* value) {
	__dpmi_regs r;
	r.d.eax = 0x0000b10a;
	r.d.ebx = bus_device_func;
	r.d.edi = reg;
	__dpmi_int(0x1a, &r);
	if (r.h.ah != 0)
		return -1;
	*value = r.d.ecx;
	return 0;
}

int pci_write_dword(unsigned bus_device_func, unsigned reg, DWORD value) {
	__dpmi_regs r;
	r.d.eax = 0x0000b10d;
	r.d.ebx = bus_device_func;
	r.d.edi = reg;
	r.d.ecx = value;
	__dpmi_int(0x1a, &r);
	if (r.h.ah != 0)
		return -1;
	return 0;
}

#ifndef __PCIREAL__
int pci_scan_device(void (*callback)(unsigned bus_device_func,unsigned vendor,unsigned device)) {
	__dpmi_regs r;
	unsigned bus;
	unsigned i,j;

	r.d.eax = 0x0000b101;
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if (r.d.edx != 0x20494350)
		return -1;

	bus = r.h.cl + 1;

	for(i=0;i<bus;++i) {
		for(j=0;j<32;++j) {
			DWORD dw;
			unsigned bus_device_func = (i << 8) | (j << 3);
			unsigned device;
			unsigned vendor;

			if (pci_read_dword(bus_device_func,0,&dw)!=0)
				continue;
			vendor = dw & 0xFFFF;
			device = (dw >> 16) & 0xFFFF;

			callback(bus_device_func,vendor,device);
		}
	}
	return 0;
}

#endif

/***************************************************************************/
/* BIOS */

/* pci BIOS address */
static unsigned pci_BIOS_selector;
static unsigned long pci_BIOS_linear_address;

/* Map the specified address */
static int pci_BIOS_address_map_plain(unsigned long phys_address) {
	if (card_create_linear_mapping(&pci_BIOS_linear_address, phys_address, 0x10000) != 0) {
		return -1;
	}
	if (card_create_selector(&pci_BIOS_selector, pci_BIOS_linear_address, 0x10000) != 0) {
		card_remove_linear_mapping(&pci_BIOS_linear_address);
		return -1;
	}
	return 0;
}

/* Map the specified address and check the signature */
static int pci_BIOS_address_map_check(unsigned long phys_address) {
	unsigned char tmp[2];
	CARD_LOG(("pci: check bios addr %08lx\n",(unsigned long)phys_address));
	if (!phys_address) {
		CARD_LOG(("pci: null bios pointer\n"));
		return -1;
	}

	tmp[0] = 0;
	tmp[1] = 0;

	if (pci_BIOS_address_map_plain(phys_address)!=0) {
		CARD_LOG(("pci: error mapping bios\n"));
		return -1;
	}

	pci_BIOS_read(tmp,0,2);

	if (tmp[0]!=0x55 || tmp[1]!=0xAA) {
		pci_BIOS_address_unmap();
		CARD_LOG(("pci: invalid bios signature\n"));
		return -1;
	}

	CARD_LOG(("pci: bios found at %08lx\n",(unsigned long)phys_address));
	return 0;
}

/* Detect and map the BIOS address */
int pci_BIOS_address_map(unsigned bus_device_func) {
#ifdef PCI_ADDRESS_REMAP_TO_1M
	DWORD addr;
	(void)bus_device_func;

	addr = 0x000c0000;
	if (!pci_BIOS_address_map_check(addr)) {
		CARD_LOG(("pci: bios not found\n"));
		return -1;
	}
#else
	DWORD orig;
	DWORD addr;
	if (pci_read_dword(bus_device_func,0x30,&orig) != 0)
		return -1;

	CARD_LOG(("pci: bios reported addr %08lx\n",(unsigned long)orig));
	addr = orig & 0xfffe0000;

	if (pci_BIOS_address_map_check(addr)!=0) {
		addr = 0x000c0000;
		if (pci_BIOS_address_map_check(addr)!=0) {
			CARD_LOG(("pci: bios not found\n"));
			return -1;
		}
	}
#endif
	return 0;
}

void pci_BIOS_address_unmap(void) {
	card_remove_selector(&pci_BIOS_selector);
	card_remove_linear_mapping(&pci_BIOS_linear_address);
	pci_BIOS_selector = 0;
}

void pci_BIOS_read(void STACK_PTR* dest, unsigned offset, unsigned len) {
	if (!pci_BIOS_selector)
		CARD_LOG(("pci: BUG! BIOS selector out of order\n"));
#ifdef __PCIREAL__
	memcpy(dest,MK_FP(pci_BIOS_selector,offset),len);
#else
	movedata(pci_BIOS_selector,offset,_my_ds(),(unsigned long)dest,len);
#endif
}

/***************************************************************************/
/* MMIO */

#ifdef PCI_ADDRESS_REMAP_TO_1M
/* Where the MMIO address are remapped */
#define PCI_MMIO_REMAP_ADDRESS 0xA0000
static DWORD pci_MMIO_orig;
#endif

/* pci MMIO address */
static unsigned pci_MMIO_selector;
static DWORD pci_MMIO_linear_address;
static unsigned pci_MMIO_register;
static unsigned pci_MMIO_bus_device_func;

int pci_MMIO_address_map(unsigned bus_device_func, unsigned reg, DWORD mask) {
	DWORD addr;
	DWORD orig;
	pci_MMIO_register = reg;
	pci_MMIO_bus_device_func = bus_device_func;
#ifdef PCI_ADDRESS_REMAP_TO_1M
	orig =	PCI_MMIO_REMAP_ADDRESS;
#else
	if (pci_read_dword(pci_MMIO_bus_device_func,pci_MMIO_register,&orig) != 0)
		return -1;
	if (!orig) {
		CARD_LOG(("pci: MMIO address null\n"));
		return -1;
	}
#endif
	CARD_LOG(("pci: MMIO orig %08lx\n",(unsigned long)orig));
	addr = orig & (0xFFFFFFF0 & mask);
	CARD_LOG(("pci: MMIO addr %08lx\n",(unsigned long)addr));
	if (card_create_linear_mapping(&pci_MMIO_linear_address, addr, 0x10000) != 0) {
		return -1;
	}
	if (card_create_selector(&pci_MMIO_selector, pci_MMIO_linear_address, 0x10000) != 0) {
		card_remove_linear_mapping(&pci_MMIO_linear_address);
		return -1;
	}
	return 0;
}

void pci_MMIO_address_unmap(void) {
	card_remove_selector(&pci_MMIO_selector);
	card_remove_linear_mapping(&pci_MMIO_linear_address);
	pci_MMIO_selector = 0;
}


int pci_MMIO_address_relocate(void) {
#ifdef PCI_ADDRESS_REMAP_TO_1M
	DWORD remap;
	DWORD check;
	if (pci_MMIO_orig)
		CARD_LOG(("pci: BUG! MMIO relocate out of order\n"));
	if (pci_read_dword(pci_MMIO_bus_device_func,pci_MMIO_register,&pci_MMIO_orig) != 0)
		return -1;
	CARD_LOG(("pci: MMIO orig %08lx\n",(unsigned long)pci_MMIO_orig));
	remap =	PCI_MMIO_REMAP_ADDRESS | (pci_MMIO_orig & 0xF);
	CARD_LOG(("pci: MMIO relocate %08lx\n",(unsigned long)remap));
	if (pci_write_dword(pci_MMIO_bus_device_func,pci_MMIO_register,remap) != 0)
		return -1;
	if (pci_read_dword(pci_MMIO_bus_device_func,pci_MMIO_register,&check) != 0)
		return -1;
	CARD_LOG(("pci: MMIO check %08lx\n",(unsigned long)check));
	if (check != remap) {
		pci_write_dword(pci_MMIO_bus_device_func,pci_MMIO_register,pci_MMIO_orig);
		return -1;
	}
#endif
	return 0;
}

void pci_MMIO_address_restore(void) {
#ifdef PCI_ADDRESS_REMAP_TO_1M
	if (!pci_MMIO_orig)
		CARD_LOG(("pci: BUG! MMIO restore out of order\n"));
	CARD_LOG(("pci: MMIO restore %08lx\n",(unsigned long)pci_MMIO_orig));
	pci_write_dword(pci_MMIO_bus_device_func,pci_MMIO_register,pci_MMIO_orig);
	pci_MMIO_orig = 0;
#endif
}

unsigned pci_MMIO_selector_get(void) {
	if (!pci_MMIO_selector)
		CARD_LOG(("pci: BUG! MMIO selector out of order\n"));
#ifdef PCI_ADDRESS_REMAP_TO_1M
	if (!pci_MMIO_orig)
		CARD_LOG(("pci: BUG! MMIO orig out of order\n"));
#endif
	return pci_MMIO_selector;
}
